from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    AnnualWellnessVisit, Ethnicity, HomeHealthcareServices, LimitedLifeExpectancy,
    OccupationalTherapyEvaluation, OfficeVisit, OncAdministrativeSex, OphthalmologicalServices,
    Payer, PhysicalTherapyEvaluation, PreventiveCareServicesEstablishedOfficeVisit18AndUp,
    PreventiveCareServicesGroupCounseling, PreventiveCareServicesIndividualCounseling,
    PreventiveCareServicesInitialOfficeVisit18AndUp, PreventiveCareServicesOther,
    PsychVisitDiagnosticEvaluation, PsychVisitPsychotherapy, Psychoanalysis, Race,
    SpeechAndHearingEvaluation, TobaccoUseCessationCounseling, TobaccoUseCessationPharmacotherapy,
    TobaccoUseScreening)


class ClinicalQualityMeasure138v9(ClinicalQualityMeasure):
    """
    Preventive Care and Screening: Tobacco Use: Screening and Cessation Intervention

    Description: Percentage of patients aged 18 years and older who were screened for tobacco use
    one or more times within 12 months AND who received tobacco cessation intervention if
    identified as a tobacco user

    Three rates are reported:
    a. Percentage of patients aged 18 years and older who were screened for tobacco use one or more
    times within 12 months
    b. Percentage of patients aged 18 years and older who were identified as a tobacco user who
    received tobacco cessation intervention
    c. Percentage of patients aged 18 years and older who were screened for tobacco use one or more
    times within 12 months AND who received tobacco cessation intervention if identified as a
    tobacco user

    Definition: Tobacco Use - Includes any type of tobacco
    Tobacco Cessation Intervention - Includes brief counseling (3 minutes or less), and/or
    pharmacotherapy -- Note: Concepts aligned with brief counseling (e.g., minimal and intensive
    advice/counseling interventions conducted both in person and over the phone) are included in
    the value set for the numerator. Other concepts such as written self-help materials (e.g.,
    brochures, pamphlets) and complementary/alternative therapies are not included in the value set
    and do not qualify for the numerator. Brief counseling also may be of longer duration or be
    performed more frequently, as evidence shows there is a dose-response relationship between the
    intensity of counseling provided (either length or frequency) and tobacco cessation rates (U.S.
    Preventive Services Task Force, 2015).

    Rationale: This measure is intended to promote adult tobacco screening and tobacco cessation
    interventions for those who use tobacco products. There is good evidence that tobacco screening
    and brief cessation intervention (including counseling and/or pharmacotherapy) is successful in
    helping tobacco users quit. Tobacco users who are able to stop using tobacco lower their risk
    for heart disease, lung disease, and stroke.

    Guidance: The requirement of two or more visits is to establish that the eligible professional
    or eligible clinician has an existing relationship with the patient for certain types of
    encounters.

    To satisfy the intent of this measure, a patient must have at least one tobacco use screening
    during the 12-month period. If a patient has multiple tobacco use screenings during the
    12-month period, only the most recent screening, which has a documented status of tobacco user
    or tobacco non-user, will be used to satisfy the measure requirements.

    If a patient uses any type of tobacco (i.e., smokes or uses smokeless tobacco), the expectation
    is that they should receive tobacco cessation intervention: either counseling and/or
    pharmacotherapy.

    As noted above in a recommendation statement from the USPSTF, the current evidence is
    insufficient to recommend electronic nicotine delivery systems (ENDS) including electronic
    cigarettes for tobacco cessation. Additionally, ENDS are not currently classified as tobacco in
    the recent evidence review to support the update of the USPSTF recommendation given that the
    devices do not burn or use tobacco leaves. In light of the current lack of evidence, the
    measure does not currently capture e-cigarette usage as either tobacco use or a cessation aid.

    If tobacco use status of a patient is unknown, the patient does not meet the screening
    component required to be counted in the numerator and should be considered a measure failure.
    Instances where tobacco use status of "unknown" is recorded include: 1) the patient was not
    screened; or 2) the patient was screened and the patient (or caregiver) was unable to provide a
    definitive answer. If the patient does not meet the screening component of the numerator but
    has an allowable medical exception, then the patient should be removed from the denominator of
    the measure and reported as a valid exception.

    In order to promote a team-based approach to patient care, the tobacco cessation intervention
    can be performed by another healthcare provider; therefore, the tobacco use screening and
    tobacco cessation intervention do not need to be performed by the same provider or clinician.

    The medical reason exception may be applied to either the screening data element OR to any of
    the applicable tobacco cessation intervention data elements (counseling and/or pharmacotherapy)
    included in the measure.

    If a patient has a diagnosis of limited life expectancy, that patient has a valid denominator
    exception for not being screened for tobacco use or for not receiving tobacco use cessation
    intervention (counseling and/or pharmacotherapy) if identified as a tobacco user.

    This measure contains three reporting rates which aim to identify patients who were screened
    for tobacco use (rate/population 1), patients who were identified as tobacco users and who
    received tobacco cessation intervention (rate/population 2), and a comprehensive look at the
    overall performance on tobacco screening and cessation intervention (rate/population 3). By
    separating this measure into various reporting rates, the eligible professional or eligible
    clinician will be able to better ascertain where gaps in performance exist, and identify
    opportunities for improvement. The overall rate (rate/population 3) can be utilized to compare
    performance to published versions of this measure prior to the 2018 performance year, when the
    measure had a single performance rate. For accountability reporting in the CMS MIPS program,
    the rate for population 2 is used for performance.

    The denominator of population criteria 2 is a subset of the resulting numerator for population
    criteria 1, as population criteria 2 is limited to assessing if patients identified as tobacco
    users received an appropriate tobacco cessation intervention. For all patients, population
    criteria 1 and 3 are applicable, but population criteria 2 will only be applicable for those
    patients who are identified as tobacco users. Therefore, data for every patient that meets the
    initial population criteria will only be submitted for population 1 and 3, whereas data
    submitted for population 2 will be for a subset of patients who meet the initial population
    criteria, as the denominator has been further limited to those who were identified as tobacco
    users.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5.  Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms138v9
    """

    title = 'Preventive Care and Screening: Tobacco Use: Screening and Cessation Intervention'

    identifiers = ['CMS138v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'American Medical Association (AMA)',
        'PCPI(R) Foundation (PCPI[R])',
    ]

    references = [
        'Siu, A. L.; U.S. Preventive Services Task Force. (2015). Behavioral and pharmacotherapy interventions for tobacco smoking cessation in adults, including pregnant women: U.S. Preventive Services Task Force recommendation statement. Annals of Internal Medicine, 163(8), 622-634. doi:10.7326/M15-2023',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: All patients aged 18 years and older seen for at least two visits or at
        least one preventive visit during the measurement period
        """
        pass

    def in_denominator(self):
        """
        Denominator: Population 1:
        Equals Initial Population

        Population 2:
        Equals Initial Population who were screened for tobacco use and identified as a tobacco
        user

        Population 3:
        Equals Initial Population

        Exclusions: None

        Exceptions: Population 1:
        Documentation of medical reason(s) for not screening for tobacco use (e.g., limited life
        expectancy, other medical reason)

        Population 2:
        Documentation of medical reason(s) for not providing tobacco cessation intervention (e.g.,
        limited life expectancy, other medical reason)

        Population 3:
        Documentation of medical reason(s) for not screening for tobacco use OR for not providing
        tobacco cessation intervention for patients identified as tobacco users (e.g., limited life
        expectancy, other medical reason)
        """
        pass

    def in_numerator(self):
        """
        Numerator: Population 1:
        Patients who were screened for tobacco use at least once within 12 months

        Population 2:
        Patients who received tobacco cessation intervention

        Population 3:
        Patients who were screened for tobacco use at least once within 12 months AND who received
        tobacco cessation intervention if identified as a tobacco user

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: The USPSTF recommends that clinicians ask all adults about tobacco
        use, advise them to stop using tobacco, and provide behavioral interventions and U.S. Food
        and Drug Administration (FDA)-approved pharmacotherapy for cessation to adults who use
        tobacco (Grade A Recommendation) (U.S. Preventive Services Task Force, 2015).

        The USPSTF recommends that clinicians ask all pregnant women about tobacco use, advise them
        to stop using tobacco, and provide behavioral interventions for cessation to pregnant women
        who use tobacco (Grade A Recommendation) (U.S. Preventive Services Task Force, 2015).

        The USPSTF concludes that the current evidence is insufficient to recommend electronic
        nicotine delivery systems for tobacco cessation in adults, including pregnant women. The
        USPSTF recommends that clinicians direct patients who smoke tobacco to other cessation
        interventions with established effectiveness and safety (previously stated) (Grade I
        Statement) (U.S. Preventive Services Task Force, 2015).
        """
        pass



from canvas_workflow_kit import events
from canvas_workflow_kit.protocol import ClinicalQualityMeasure
from canvas_workflow_kit.timeframe import Timeframe
from canvas_workflow_kit.recommendation import ImmunizationRecommendation
from canvas_workflow_kit.value_set.v2018 import (
    AllergyToEggs, AllergyToInfluenzaVaccine, AnnualWellnessVisit,
    CareServicesInLongTermResidentialFacility, DischargeServicesNursingFacility, EggSubstance,
    EncounterInfluenza, Ethnicity, FaceToFaceInteraction, Hemodialysis, HomeHealthcareServices,
    InfluenzaVaccination, InfluenzaVaccinationDeclined, InfluenzaVaccine_1254,
    IntoleranceToInfluenzaVaccine, NursingFacilityVisit, OfficeVisit, OncAdministrativeSex,
    OutpatientConsultation, PatientProviderInteraction, Payer, PeritonealDialysis,
    PreventiveCareEstablishedOfficeVisit0To17, PreventiveCareInitialOfficeVisit0To17,
    PreventiveCareServicesEstablishedOfficeVisit18AndUp, PreventiveCareServicesGroupCounseling,
    PreventiveCareServicesIndividualCounseling, PreventiveCareServicesInitialOfficeVisit18AndUp,
    PreventiveCareServicesOther, PreviousReceiptOfInfluenzaVaccine, Race)  # flake8: noqa


class ClinicalQualityMeasure147v7(ClinicalQualityMeasure):
    """
    Preventive Care and Screening: Influenza Immunization

    Description: Percentage of patients aged 6 months and older seen for a visit between October 1
    and March 31 who received an influenza immunization OR who reported previous receipt of an
    influenza immunization

    Definition: Previous Receipt - receipt of the current season's influenza immunization from
    another provider OR from same provider prior to the visit to which the measure is applied
    (typically, prior vaccination would include influenza vaccine given since August 1st)

    Rationale: Influenza vaccination is the most effective protection against influenza virus
    infection (CDC, 2016). Influenza may lead to serious complications including hospitalization or
    death (CDC, 2016). Influenza vaccine is recommended for all persons aged >=6 months who do not
    have contraindications to vaccination. However, data indicate that less than half of all
    eligible individuals receive an influenza vaccination (CDC, 2015). This measure promotes annual
    influenza vaccination for all persons aged >= 6 months.

    Guidance: The timeframe for the visit during the "Encounter, Performed: Encounter-Influenza" or
    "Procedure, Performed: Peritoneal Dialysis" or "Procedure, Performed: Hemodialysis" in the
    Population Criteria-Denominator, refers to the influenza season defined by the measure: October
    through March (October 1 for the year prior to the start of the reporting period through March
    31 during the reporting period).  The "Encounter-Influenza" Grouping OID detailed in the data
    criteria section below is comprised of several individual OIDs of different encounter types.
    The individual OIDs are included in the value set and should be reviewed to determine that an
    applicable visit occurred during the timeframe for "Encounter, Performed: Encounter-Influenza"
    as specified in the denominator.

    To enable reporting of this measure at the close of the reporting period, this measure will
    only assess the influenza season that ends in March of the reporting period. The subsequent
    influenza season (ending March of the following year) will be measured and reported in the
    following year.

    To account for the majority of reporting years' appropriate flu season duration, the measure
    logic will look at the first 89 days of the measurement period for the appropriate criteria and
    actions to be present/performed (January 1 through March 31). The measure developer believes it
    is best to keep the logic as static as possible from one reporting year to the next. Therefore,
    during leap years, only encounters that occur through March 30 will be counted in the
    denominator.

    As a result of updated CDC/ACIP guidelines which include the interim recommendation that live
    attenuated influenza vaccine (LAIV) should not be used due to low effectiveness against
    influenza A(H1N1)pdm09 in the United States during the 2013-14 and 2015-16 seasons, the measure
    specifications have been updated and no longer include LAIV or intranasal flu vaccine as an
    option for numerator eligibility.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms147v7
    """

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'American Medical Association (AMA)',
        'PCPI(R) Foundation (PCPI[R])',
    ]

    references = [
        'Seasonal Influenza: Flu Basics. Centers for Disease Control and Prevention Web site.  http://www.cdc.gov/flu/about/disease/index.htm. Updated May 4, 2016. Accessed June 23, 2016.',
        'Flu vaccination coverage: United States, 2014-15 Influenza Season. Centers for Disease Control and Prevention Web site.  http://www.cdc.gov/flu/fluvaxview/coverage-1415estimates.htm. Updated September 17, 2015. Accessed June 23, 2016.',
        'Grohskopf LA, Sokolow LZ, Broder KR, et al. Prevention and Control of Seasonal Influenza with Vaccines. MMWR Recomm Rep 2016;65(No. RR-5):1-54. DOI: http://dx.doi.org/10.15585/mmwr.rr6505a',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: All patients aged 6 months and older seen for a visit during the
        measurement period
        """
        # FIX: what exactly is 6 months old?  Probably we need age_at
        # to return days here not fractions of a year.
        age = self.patient.age_at(self.timeframe.end)
        if age < 0.5:
            return False

        return True

    def in_denominator(self):
        """
        Denominator: Equals Initial Population and seen for a visit between October 1 and March 31

        Exclusions: None

        Exceptions: Documentation of medical reason(s) for not receiving influenza immunization
        (eg, patient allergy, other medical reasons).

        Documentation of patient reason(s) for not receiving influenza immunization (eg, patient
        declined, other patient reasons).

        Documentation of system reason(s) for not receiving influenza immunization (eg, vaccine not
        available, other system reasons).
        """
        if not self.in_initial_population():
            return False

        if self.patient.conditions.before(
                self.timeframe.start).find(IntoleranceToInfluenzaVaccine):
            return False

        # FIX: how to check for patient declined flu shot?

        return True

    def immunization_timeframe(self):
        """
        Returns the timeframe within-which the immunization is recommended, Oct 1 to March 31.
        """
        start = self.timeframe.start.replace(month=10, day=1)
        end = start.shift(months=5, days=30)
        return Timeframe(start=start, end=end)

    def in_numerator(self):
        """
        Numerator: Patients who received an influenza immunization OR who reported previous receipt
        of an influenza immunization

        Exclusions: Not Applicable
        """
        if self.patient.immunizations.within(
                self.immunization_timeframe()).find(InfluenzaVaccine_1254):
            return True

        return False

    def compute_results(self):
        """
        Clinical recommendation: Routine annual influenza vaccination is recommended for all
        persons aged >= 6 months who do not have contraindications. Optimally, vaccination should
        occur before onset of influenza activity in the community. Health care providers should
        offer vaccination by October, if possible. Vaccination should continue to be offered as
        long as influenza viruses are circulating. (CDC/Advisory Committee on Immunization
        Practices (ACIP), 2016)
        """
        patient = self.patient
        recommendations = []
        status = 'not_applicable'
        narrative = ''

        # only recommend flu vaccines between October 1 and the end of March
        immunization_timeframe = self.immunization_timeframe()
        if immunization_timeframe.contains(self.timeframe.start):
            rec = ImmunizationRecommendation(patient, InfluenzaVaccine_1254)
            recommendations.append(rec)

            # FIX: check if they already have the vaccine
            if self.patient.immunizations.within(immunization_timeframe).find(
                    InfluenzaVaccine_1254):
                status = 'current'
                narrative = f'{patient.first_name} has their {InfluenzaVaccine_1254.name} this year.'
            else:
                status = 'due'
                narrative = rec.narrative

        return {'status': status, 'narrative': narrative, 'recommendations': recommendations}



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    AcuteInpatient, AnnualWellnessVisit, CareServicesInLongTermResidentialFacility, CtColonography,
    DementiaMedications, Ed, EncounterInpatient, Ethnicity, FecalOccultBloodTestFobt, FitDna,
    FrailtyDevice, FrailtyDiagnosis, FrailtyEncounter, FrailtySymptom, HomeHealthcareServices,
    HospiceCareAmbulatory, MalignantNeoplasmOfColon, NonacuteInpatient, NursingFacilityVisit,
    Observation, OfficeVisit, OncAdministrativeSex, Outpatient, Payer,
    PreventiveCareServicesEstablishedOfficeVisit18AndUp,
    PreventiveCareServicesInitialOfficeVisit18AndUp, Race)


class ClinicalQualityMeasure130v9(ClinicalQualityMeasure):
    """
    Colorectal Cancer Screening

    Description: Percentage of adults 50-75 years of age who had appropriate screening for
    colorectal cancer

    Definition: None

    Rationale: Colorectal cancer represents eight percent of all new cancer cases and is the second
    leading cause of cancer deaths in the United States. In 2019, an estimated 145,600 new cases of
    colorectal cancer and an estimated 51,020 deaths attributed to it. According to the National
    Cancer Institute, about 4.2 percent of men and women will be diagnosed with colorectal cancer
    at some point during their lifetimes. For most adults, older age is the most important risk
    factor for colorectal cancer, although being male and black are also associated with higher
    incidence and mortality. Colorectal cancer is most frequently diagnosed among people 65 to 74
    years old (National Cancer Institute, 2019).

    Screening can be effective for finding precancerous lesions (polyps) that could later become
    malignant, and for detecting early cancers that can be more easily and effectively treated.
    Precancerous polyps usually take about 10 to 15 years to develop into colorectal cancer, and
    most can be found and removed before turning into cancer. The five-year relative survival rate
    for people whose colorectal cancer is found in the early stage before it has spread is about 90
    percent (American Cancer Society, 2018).

    Guidance: Patient self-report for procedures as well as diagnostic studies should be recorded
    in "Procedure, Performed" template or "Diagnostic Study, Performed" template in QRDA-1.

    Do not count digital rectal exams (DRE), fecal occult blood tests (FOBTs) performed in an
    office setting or performed on a sample collected via DRE.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5.  Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms130v9
    """

    title = 'Colorectal Cancer Screening'

    identifiers = ['CMS130v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'National Committee for Quality Assurance',
    ]

    references = [
        'American Cancer Society. (2018). Can Colorectal Polyps and Cancer Be Found Early? Retrieved October 17, 2019, from https://www.cancer.org/cancer/colon-rectal-cancer/detection-diagnosis-staging/detection.html',
        'National Cancer Institute. (2019). SEER Cancer Statistics Review, 1975-2016. Retrieved October 17, 2019, from SEER website: https://seer.cancer.gov/csr/1975_2016/index.html',
        'US Preventive Services Task Force, Bibbins-Domingo, K., Grossman, D. C., Curry, S. J., Davidson, K. W., Epling, J. W., … Siu, A. L. (2016). Screening for Colorectal Cancer: US Preventive Services Task Force Recommendation Statement. JAMA, 315(23), 2564–2575. https://doi.org/10.1001/jama.2016.5989',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Patients 50-75 years of age with a visit during the measurement period
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Exclude patients whose hospice care overlaps the measurement period.

        Exclude patients with a diagnosis or past history of total colectomy or colorectal cancer.

        Exclude patients 66 and older who are living long term in an institution for more than 90
        consecutive days during the measurement period.

        Exclude patients 66 and older with advanced illness and frailty because it is unlikely that
        patients will benefit from the services being measured.

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Patients with one or more screenings for colorectal cancer. Appropriate
        screenings are defined by any one of the following criteria:
        - Fecal occult blood test (FOBT) during the measurement period
        - Flexible sigmoidoscopy during the measurement period or the four years prior to the
        measurement period
        - Colonoscopy during the measurement period or the nine years prior to the measurement
        period
        - FIT-DNA during the measurement period or the two years prior to the measurement period
        - CT Colonography during the measurement period or the four years prior to the measurement
        period

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: The U.S. Preventive Services Task Force (2016) recommends
        screening for colorectal cancer starting at age 50 years and continuing until age 75 years.
        This is a Grade A recommendation (U.S. Preventive Services Task Force, 2016).
        Appropriate screenings are defined by any one of the following:
        -Colonoscopy (every 10 years)
        -Flexible sigmoidoscopy (every 5 years)
        -Fecal occult blood test (annually)
        -FIT-DNA (every 3 years)
        -Computed tomographic colonography (every 5 years)
        """
        pass



from canvas_workflow_kit.value_set.value_set import ValueSet
from canvas_workflow_kit.value_set.v2020 import KidneyFailure as v2020KidneyFailure


class KidneyFailure(v2020KidneyFailure):
    """
    Added N1830-N1832 per MRHS request
    """

    OID = '2.16.840.1.113883.3.464.1003.109.12.1028'
    VALUE_SET_NAME = 'Kidney Failure'
    EXPANSION_VERSION = 'eCQM Update 2019-05-10 with a few additions'

    ICD10CM = {
        'N170', 'N171', 'N172', 'N178', 'N179', 'N181', 'N182', 'N183', 'N184', 'N185', 'N186',
        'N189', 'N19', 'N1830', 'N1831', 'N1832'
    }


# @canvas-adr-0006
class Hcc005v1AnnualWellnessVisit(ValueSet):
    VALUE_SET_NAME = 'Annual Wellness Visit'
    EXPANSION_VERSION = 'CanvasHCC Update 2019-11-04'

    HCPCS = {
        'G0438',
        'G0439',
        'G0402',
        '99387',
        '99397',
    }


# @canvas-adr-0006
class CMS125v6Tomography(ValueSet):
    VALUE_SET_NAME = 'Tomography'
    EXPANSION_VERSION = 'Update 2019-08-01'

    LOINC = {
        '72142-3',
    }


# @canvas-adr-0006
class CMS130v6CtColonography(ValueSet):
    VALUE_SET_NAME = 'Ct Colonography'
    EXPANSION_VERSION = 'Update 2019-08-01'

    LOINC = {
        '79101-2',
    }


class CMS134v6Dialysis(ValueSet):
    VALUE_SET_NAME = 'CMS 134v6 Dialysis'
    EXPANSION_VERSION = 'CanvasHCC Update 2018-12-05'

    ICD10CM = {
        'Z992',
    }

    SNOMEDCT = {
        '207RN0300X',
        '2080P0210X',
    }


class LabReportCreatinine(ValueSet):
    VALUE_SET_NAME = 'Lab Report Creatinine'
    EXPANSION_VERSION = 'CanvasHCC Update 2018-10-04'

    LOINC = {
        '2160-0',
    }


class DiabetesWithoutComplication(ValueSet):
    VALUE_SET_NAME = 'Diabetes Without Complication'
    EXPANSION_VERSION = 'CanvasHCC Update 2018-10-16'

    ICD10CM = {
        'E119',
    }


class DiabetesEyeConditionSuspect(ValueSet):
    VALUE_SET_NAME = 'Diabetes Eye Condition suspect'
    EXPANSION_VERSION = 'CanvasHCC Update 2018-10-16'

    ICD10CM = {
        'H28',
        'H36',
    }


class DiabetesEyeClassConditionSuspect(ValueSet):
    VALUE_SET_NAME = 'Diabetes Eye Class Condition suspect'
    EXPANSION_VERSION = 'CanvasHCC Update 2018-10-16'

    ICD10CM = {
        'H35\d*',
    }


class DiabetesNeurologicConditionSuspect(ValueSet):
    VALUE_SET_NAME = 'Diabetes Neurologic Condition suspect'
    EXPANSION_VERSION = 'CanvasHCC Update 2018-10-16'

    ICD10CM = {
        'G63',
        'G737',
        'G53',
    }


class DiabetesRenalConditionSuspect(ValueSet):
    VALUE_SET_NAME = 'Diabetes Renal Condition suspect'
    EXPANSION_VERSION = 'CanvasHCC Update 2018-10-16'

    ICD10CM = {
        'N181',
        'N182',
        'N183',
        'N184',
        'N185',
        'N186',
        'N189',
    }


class DiabetesCirculatoryClassConditionSuspect(ValueSet):
    VALUE_SET_NAME = 'Diabetes Circulatory Class Condition suspect'
    EXPANSION_VERSION = 'CanvasHCC Update 2018-10-16'

    ICD10CM = {
        'I73\d*',
        'I70\d*',
        'I71\d*',
        'I79\d*',
    }


class DiabetesOtherClassConditionSuspect(ValueSet):
    VALUE_SET_NAME = 'Diabetes Other Class Condition suspect'
    EXPANSION_VERSION = 'CanvasHCC Update 2018-10-16'

    ICD10CM = {
        'M14\d*',
        'L97\d*',
        'L984\d*',
    }


class DysrhythmiaClassConditionSuspect(ValueSet):
    VALUE_SET_NAME = 'Dysrhythmia Class Condition suspect'
    EXPANSION_VERSION = 'CanvasHCC Update 2018-10-18'

    ICD10CM = {
        'I42\d*',
        'I47\d*',
        'I48\d*',
        'I49\d*',
    }


class Covid19QuestionnaireSymptomaticSurveillance(ValueSet):
    VALUE_SET_NAME = 'Covid19 Questionnaire Symptomatic Surveillance'
    EXPANSION_VERSION = 'CanvasCCP Update 2020-03-19'

    CANVAS = {
        'CANVAS0001',
    }


class Covid19QuestionnaireHighRiskOutreach(ValueSet):
    VALUE_SET_NAME = 'Covid19 Questionnaire High Risk Outreach'
    EXPANSION_VERSION = 'CanvasCCP Update 2020-03-24'

    CANVAS = {
        'CANVAS0006',
    }



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    BoneScan, Ethnicity, Male, OncAdministrativeSex, PainRelatedToProstateCancer, Payer,
    ProstateCancer, ProstateSpecificAntigenTest, Race)


class ClinicalQualityMeasure129v10(ClinicalQualityMeasure):
    """
    Prostate Cancer: Avoidance of Overuse of Bone Scan for Staging Low Risk Prostate Cancer
    Patients

    Description: Percentage of patients, regardless of age, with a diagnosis of prostate cancer at
    low (or very low) risk of recurrence receiving interstitial prostate brachytherapy, OR external
    beam radiotherapy to the prostate, OR radical prostatectomy who did not have a bone scan
    performed at any time since diagnosis of prostate cancer

    Definition: Risk Strata Definitions: Very Low, Low, Intermediate, High, or Very High-
    Very Low/Low Risk - PSA < 10 ng/mL; AND Gleason score 6 or less/Gleason grade group 1; AND
    clinical stage T1 to T2a.
    Intermediate Risk - PSA 10 to 20 ng/mL; OR Gleason score 7/Gleason grade group 2-3; OR clinical
    stage T2b to T2c.
    High/Very High Risk - PSA > 20 ng/mL; OR Gleason score 8 to 10/Gleason grade group 4-5; OR
    clinically localized stage T3 to T4 (adapted from the National Comprehensive Cancer Network,
    2018).

    External beam radiotherapy - external beam radiotherapy refers to 3D conformal radiation
    therapy (3D-CRT), intensity modulated radiation therapy (IMRT), stereotactic body radiotherapy
    (SBRT), and proton beam therapy.

    Bone scan - bone scan refers to the conventional technetium-99m-MDP bone scan as well as
    18F-NaF PET (or PET/CT) scan.

    Rationale: Multiple studies have indicated that a bone scan is not clinically necessary for
    staging prostate cancer in men with a low (or very low) risk of recurrence and receiving
    primary therapy. For patients who are categorized as low-risk, bone scans are unlikely to
    identify their disease. Furthermore, bone scans are not necessary for low-risk patients who
    have no history or if the clinical examination suggests no bony involvement. Less than 1% of
    low-risk patients are at risk of metastatic disease.

    While clinical practice guidelines do not recommend bone scans in low-risk prostate cancer
    patients, overuse is still common. An analysis of prostate cancer patients in the SEER-Medicare
    database diagnosed from 2004-2007 found that 43% of patients for whom a bone scan was not
    recommended received it (Falchook, Hendrix, & Chen, 2015). The analysis also found that the use
    of bone scans in low-risk patients leads to an annual cost of $4 million dollars to Medicare.
    The overuse of bone scan imaging for low-risk prostate cancer patients is a concept included on
    the American Urological Association's (AUA) list in the Choosing Wisely Initiative as a means
    to promote adherence to evidence-based imaging practices and to reduce health care dollars
    wasted (AUA, 2017). This measure is intended to promote adherence to evidence-based imaging
    practices, lessen the financial burden of unnecessary imaging, and ultimately to improve the
    quality of care for prostate cancer patients in the United States.

    Guidance: A higher score indicates appropriate treatment of patients with prostate cancer at
    low (or very low) risk of recurrence. Only patients with prostate cancer with low (or very low)
    risk of recurrence will be counted in the performance denominator of this measure.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5. Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms129v10
    """

    title = 'Prostate Cancer: Avoidance of Overuse of Bone Scan for Staging Low Risk Prostate Cancer Patients'

    identifiers = ['CMS129v10']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'American Medical Association (AMA)',
        'PCPI(R) Foundation (PCPI[R])',
    ]

    references = [
        'American Urological Association. (2017). A routine bone scan is unnecessary in men with low-risk prostate cancer. Retrieved from http://www.choosingwisely.org/clinician-lists/american-urological-association-routine-bone-scans-with-low-risk-prostate-cancer/ (Original work published in 2013).',
        'American Urological Association, American Society for Radiation Oncology, & Society of Urologic Oncology. (2017). Clinically localized prostate cancer: AUA/ASTRO/SUO Guideline. Retrieved from https://www.auanet.org/guidelines/prostate-cancer-clinically-localized-(2017)',
        'Falchook, A. D., Hendrix, L. H., & Chen, R. C. (2015). Guideline-discordant use of imaging during work-up of newly diagnosed prostate cancer. Journal of Oncology Practice, 11(2), e239-e246. doi:10.1200/jop.2014.001818',
        'National Comprehensive Cancer Network. (2019). Clinical Practice Guidelines in Oncology: Prostate Cancer. Version 4.209. Retrieved from https://www.nccn.org/professionals/physician_gls/pdf/prostate.pdf',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: All patients, regardless of age, with a diagnosis of prostate cancer
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population at low (or very low) risk of recurrence receiving
        interstitial prostate brachytherapy, OR external beam radiotherapy to the prostate, OR
        radical prostatectomy

        Exclusions: None

        Exceptions: Documentation of reason(s) for performing a bone scan (including documented
        pain, salvage therapy, other medical reasons, bone scan ordered by someone other than
        reporting physician)
        """
        pass

    def in_numerator(self):
        """
        Numerator: Patients who did not have a bone scan performed at any time since diagnosis of
        prostate cancer

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: For symptomatic patients and/or those with a life expectancy of
        greater than 5 years, bone imaging is appropriate for patients with unfavorable
        intermediate-risk prostate cancer and T2 disease with PSA over 10 ng/mL, high- or very-high
        risk disease; or symptomatic disease (National Comprehensive Cancer Network, 2019)
        (Evidence Level: Category 2A).

        Clinicians should not perform routine bone scans in the staging of asymptomatic very low-
        or low-risk localized prostate cancer patients (AUA, American Society for Radiation
        Oncology, & Society of Urologic Oncology, 2017) (Strong Recommendation; Evidence Level:
        Grade C).

        A routine bone scan is unnecessary in men with low-risk prostate cancer. Low-risk patients
        are unlikely to have disease identified by bone scan. Accordingly, bone scans are generally
        unnecessary in patients with newly diagnosed prostate cancer who have a PSA <10.0 ng/mL and
        a Gleason score less than 7 unless the patient’s history or clinical examination suggests
        bony involvement. Progression to the bone is much more common in advanced local disease or
        in high-grade disease that is characterized by fast and aggressive growth into surrounding
        areas such as bones or lymph nodes (AUA, 2017).
        """
        pass



from typing import Dict, List

from canvas_workflow_kit import events
from canvas_workflow_kit.canvas_code_set import CanvasCodeSet
from canvas_workflow_kit.intervention import Intervention
from canvas_workflow_kit.patient_recordset import InterviewRecordSet
from canvas_workflow_kit.protocol import STATUS_DUE, STATUS_SATISFIED, ClinicalQualityMeasure, ProtocolResult
from canvas_workflow_kit.recommendation import (
    Recommendation, ImmunizationRecommendation
)
from canvas_workflow_kit.timeframe import Timeframe
from canvas_workflow_kit.value_set.specials import (
    Covid19QuestionnaireHighRiskOutreach,
    Covid19QuestionnaireSymptomaticSurveillance
)


from canvas_workflow_kit.value_set.v2018 import (
    InfluenzaVaccine_1254
)

# flake8: noqa


class HyperlinkRecommendation(ClinicalQualityMeasure):

    class Meta:
        title = 'Hyperlink Recommendation'
        version = "1.2"
        changelog = "Initial release"

        description = 'All patients with COVID Questionnaire completed Date < 7 days ago and >  5 days ago.'
        information = 'https://canvas-medical.zendesk.com/hc/en-us/articles/360059084173-COVID-19-Risk-Assessment-Follow-Up-Protocol'

        identifiers = ['CCP001v1']

        types = ['CCP']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]
        authors = [
            'Canvas Medical Team',
        ]

        compute_on_change_types = [
            ClinicalQualityMeasure.CHANGE_INTERVIEW,
        ]

    def compute_results(self) -> ProtocolResult:
        result = ProtocolResult()

        result.add_recommendation(
            Intervention(
                title='Link Rec Title',
                narrative=f'Link Rec Narr',
                href='http://canvasmedical.com'
            )
        )
        result.add_recommendation(
            ImmunizationRecommendation(
                key='KEY-ID',
                rank=123,
                button='ACT',
                patient=self.patient,
                immunization=InfluenzaVaccine_1254)
        )

        return result



import arrow

from canvas_workflow_kit.timeframe import Timeframe
from canvas_workflow_kit.value_set.v2018 import AntibioticMedicationsForPharyngitis, ContraceptiveMedications

from .base import SDKBaseTest


class SDKMedicationTest(SDKBaseTest):

    def test_medications(self):
        patient = self.load_patient('example')
        self.assertTrue(patient)

        # look for meds he doesn't have
        t = Timeframe(arrow.get('2017-01-01'), arrow.get('2018-02-01'))
        self.assertFalse(
            patient.medications.find(ContraceptiveMedications).intersects(t, still_active=False))

        # look for a med he does have
        self.assertTrue(
            patient.medications.find(AntibioticMedicationsForPharyngitis).intersects(
                t, still_active=False))



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    AnnualWellnessVisit, AntidepressantMedication, EncounterInpatient, Ethnicity,
    HomeHealthcareServices, HospiceCareAmbulatory, MajorDepression, NursingFacilityVisit,
    OfficeVisit, OncAdministrativeSex, Payer, PreventiveCareServicesEstablishedOfficeVisit18AndUp,
    PreventiveCareServicesInitialOfficeVisit18AndUp, PsychVisitDiagnosticEvaluation,
    PsychVisitPsychotherapy, Race, TelephoneEvaluation, TelephoneManagement)


class ClinicalQualityMeasure128v9(ClinicalQualityMeasure):
    """
    Anti-depressant Medication Management

    Description: Percentage of patients 18 years of age and older who were treated with
    antidepressant medication, had a diagnosis of major depression, and who remained on an
    antidepressant medication treatment. Two rates are reported.
    a. Percentage of patients who remained on an antidepressant medication for at least 84 days (12
    weeks).
    b. Percentage of patients who remained on an antidepressant medication for at least 180 days (6
    months).

    Definition: Index Prescription Start Date (IPSD): The earliest prescription dispensing event
    for an antidepressant medication during the period of 245 days prior to the start of the
    measurement period through 120 days after the start of the measurement period.
    The "continuous treatment" described in this measure allows for gaps in medication treatment up
    to a total 30 days during the 114-day period (numerator 1) or 51 days during the 231-day period
    (numerator 2). Gaps can include either gaps used to change medication, or treatment gaps to
    refill the same medication.

    Rationale: In 2017, over 17 million adults in the United States had at least one major
    depressive episode in the past 12 months (SAMHSA, 2018), and depression is estimated to affect
    nearly a quarter of adults in their lifetime (Burcusa & Iacono, 2007). Depression is associated
    with other chronic diseases, as it adversely affects the course, complications and management
    of other chronic medical illnesses such as diabetes, cancer, cardiovascular disease and asthma
    (Katon & Guico-Pabia, 2011).

    Symptoms of depression include appetite and sleep disturbances, anxiety, irritability and
    decreased concentration (Charbonneau
    et al., 2005). The American Psychiatric Association recommends use of antidepressant medication
    and behavioral therapies, such as psychotherapy, to treat depression (American Psychiatric
    Association, 2010).

    For the past 50 years, antidepressant medication has proven to be effective especially for
    patients with more severe symptoms (Fournier et al., 2010). Among patients who initiate
    antidepressant treatment, one in three discontinues treatment within one month, before the
    effect of medication can be assessed, and nearly one in two discontinues treatment within three
    months (Simon, 2002).

    Clinical guidelines for depression emphasize the importance of effective clinical management in
    increasing patients’ medication compliance, monitoring treatment effectiveness, and identifying
    and managing side effects. If pharmacological treatment is initiated, appropriate dosing and
    continuation of therapy through the acute and continuation phases decrease recurrence of
    depression. Thus, evaluation of duration of pharmacological treatment serves as an important
    indicator in promoting patient compliance with the establishment and maintenance of an
    effective medication regimen.

    Guidance: To identify new treatment episodes for major depression, there must be a 105-day
    negative medication history (a period during which the patient was not taking antidepressant
    medication) prior to the first dispensing event associated with the Index Episode Start Date
    (Index Prescription Start Date).

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5. Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms128v9
    """

    title = 'Anti-depressant Medication Management'

    identifiers = ['CMS128v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'National Committee for Quality Assurance',
    ]

    references = [
        'American Psychiatric Association. (2010). Practice guideline for the treatment of patients with major depressive disorder, 3rd edition. Arlington: Author.',
        'Burcusa, S. L., & Iacono, W. G. (2007). Risk for recurrence in depression. Clinical Psychology Review, 27(8), 959-985.',
        'Charbonneau, A., Bruning, W., Titus-Howard, T., et al. (2005). The community initiative on depression: Report from a multiphase work site depression intervention. Journal of Occupational and Environmental Medicine, 47(1), 60-67.',
        'Fournier, J. C., DeRubeis, R. J., Hollon, S. D., et al. (2010). Antidepressant drug effects and depression severity: A patient-level meta-analysis. JAMA, 303(1), 47-53.',
        'Katon, W., & Guico-Pabia, C. J. (2011). Improving quality of depression care using organized systems of care: A review of the literature. The Primary Care Companion to CNS Disorders, 13(1).',
        'Simon, G. E. (2002). Evidence review: Efficacy and effectiveness of antidepressant treatment in primary care. General Hospital Psychiatry, 24(4), 213-224.',
        'Substance Abuse and Mental Health Services Administration (SAMHSA). (2018). Results from the 2017 National Survey on drug use and health: detailed tables. Retrieved from: https://www.samhsa.gov/data/sites/default/files/cbhsq-reports/NSDUHDetailedTabs2017/NSDUHDetailedTabs2017.htm#tab8-56A',
        'Department of Veterans Affairs, and Health Affairs, Department of Defense. (2000). Management of Major Depressive Disorder in Adults in the Primary Care Setting. Retrieved from: http://www.oqp.med.va.gov/cpg/MDD/MDD_Base.htm',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Patients 18 years of age and older who were dispensed antidepressant
        medications within 245 days (8 months) prior to the measurement period through the first
        120 days (4 months) of the measurement period, and were diagnosed with major depression 60
        days prior to, or 60 days after the dispensing event and had a visit 60 days prior to, or
        60 days after the dispensing event
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Patients who were actively on an antidepressant medication in the 105 days
        prior to the Index Prescription Start Date.

        Exclude patients whose hospice care overlaps the measurement period.

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Numerator 1: Patients who have received antidepressant medication for at least
        84 days (12 weeks) of continuous treatment during the 114-day period following the Index
        Prescription Start Date.

        Numerator 2: Patients who have received antidepressant medications for at least 180 days (6
        months) of continuous treatment during the 231-day period following the Index Prescription
        Start Date.

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: American Psychiatric Association (2010):
        • “An antidepressant medication is recommended as an initial treatment choice for patients
        with mild to moderate major depressive disorder [I: Recommended with substantial clinical
        confidence] and definitely should be provided for those with severe major depressive
        disorder unless electroconvulsive therapy (ECT) is planned [I: Recommended with substantial
        clinical confidence].”

        • Patients should be given a realistic notion of what can be expected during the different
        phases of treatment, including the likely time course of symptom response and the
        importance of adherence for successful treatment and prophylaxis [I].

        • During the acute phase of treatment, patients should be
        carefully and systematically monitored on a regular basis
        to assess their response to pharmacotherapy, identify the
        emergence of side effects (e.g., gastrointestinal symptoms,
        sedation, insomnia, activation, changes in weight, and cardiovascular,
        neurological, anticholinergic, or sexual side effects),
        and assess patient safety [I].

        • “During the continuation phase of treatment, the patient should be carefully monitored
        for signs of possible relapse [I: Recommended with substantial clinical confidence].
        Systematic assessment of symptoms, side effects, adherence, and functional status is
        essential [I: Recommended with substantial clinical confidence], and may be facilitated
        through the use of clinician- and/or patient-administered rating scales [II: Recommended
        with moderate clinical confidence]. To reduce the risk of relapse, patients who have been
        treated successfully with antidepressant medications in the acute phase should continue
        treatment with these agents for 4–9 months [I: Recommended with substantial clinical
        confidence].”

        Department of Veterans Affairs, and Health Affairs, Department of Defense (2000):
        •“When antidepressant pharmacotherapy is used, the following key messages should be given
        to enhance adherence to medication: [B: A recommendation that clinicians provide (the
        service) to eligible patients.]
        • Most people need to be on medication for at least 6 to 12 months after adequate response
        • It usually takes 2 to 6 weeks before improvements are seen
        • Continue to take the medication even after feeling better
        • Do not discontinue taking medications without first discussing with your provider.”
        """
        pass



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    EncounterInpatient, Ethnicity, HeartFailure, HospiceCareAmbulatory, OfficeVisit,
    OncAdministrativeSex, Payer, Race)


class ClinicalQualityMeasure90v10(ClinicalQualityMeasure):
    """
    Functional Status Assessments for Congestive Heart Failure

    Description: Percentage of patients 18 years of age and older with congestive heart failure who
    completed initial and follow-up patient-reported functional status assessments

    Definition: None

    Rationale: Patients living with congestive heart failure (CHF) often have poor functional
    status and health-related quality of life, which declines as the disease progresses (Allen et
    al., 2012). In addition, their care is often complicated by multiple comorbidities. To assist
    in managing these complex patients, the American College of Cardiology Foundation and American
    Heart Association recommend collecting initial and repeat assessments of a patient's function
    and ability to complete desired activities of daily living (Hunt et al., 2009). The American
    Heart Association has also released scientific statements emphasizing the collection of
    patient-reported health status (for example, functional limitations, symptom burden, quality of
    life) from CHF patients as an important means of establishing a dynamic conversation between
    patient and provider regarding care goals and the patient's priorities (Allen et al., 2012;
    Rumsfeld et al., 2013).

    Guidance: Initial functional status assessment (FSA) and encounter: The initial FSA is the
    first FSA that occurs two weeks before or during the first encounter in the first 185 days of
    the measurement year.

    Follow-up FSA: The follow-up FSA must be completed at least 30 days but no more than 180 days
    after the initial FSA.

    The same FSA instrument must be used for the initial and follow-up assessment.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5.  Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms90v10
    """

    title = 'Functional Status Assessments for Congestive Heart Failure'

    identifiers = ['CMS90v10']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'National Committee for Quality Assurance',
    ]

    references = [
        'Allen, L. A., Stevenson, L. W., Grady, K. L., et al. (2012). Decision making in advanced heart failure: A scientific statement from the American Heart Association. Circulation, 125(15), 1928-1952. doi: 10.1161/CIR.0b013e31824f2173',
        'American College of Cardiology Foundation & American Heart Association. (2013). Guideline for the management of heart failure: A report of the American College of Cardiology Foundation/American Heart Association Task Force on Practice Guidelines. Circulation, 128(16), e240-e327. doi: 10.1161/CIR.0b013e31829e8776',
        'Hunt, S. A., Abraham, W. T., Chin, M. H., et al. (2009). 2009 focused update incorporated into the ACC/AHA 2005 guidelines for the diagnosis and management of heart failure in adults. Circulation, 119(14), e391-e479. doi: 10.1161/CIRCULATIONAHA.109.192065',
        'Rumsfeld, J. S., Alexander, K. P., Goff, D. C., et al. (2013). Cardiovascular health: The importance of measuring patient-reported health status: A scientific statement from the American Heart Association. Circulation, 127(22), 2233-2249. doi: 10.1161/CIR.0b013e3182949a2e',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Patients 18 years of age and older who had two outpatient encounters
        during the measurement year and a diagnosis of congestive heart failure
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Exclude patients with severe cognitive impairment that overlaps the measurement
        period.

        Exclude patients whose hospice care overlaps the measurement period.

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Patients with patient-reported functional status assessment  results (i.e.,
        Veterans RAND 12-item health survey [VR-12]; VR-36; Kansas City Cardiomyopathy
        Questionnaire [KCCQ]; KCCQ-12; Minnesota Living with Heart Failure Questionnaire [MLHFQ];
        Patient-Reported Outcomes Measurement Information System [PROMIS]-10 Global Health,
        PROMIS-29) present in the EHR two weeks before or during the initial FSA encounter and
        results for the follow-up FSA at least 30 days but no more than 180 days after the initial
        FSA

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: American College of Cardiology Foundation and American Heart
        Association (2013): While this guideline does not explicitly recommend the use of patient-
        reported functional status or quality of life assessments (such as the Kansas City
        Cardiomyopathy Questionnaire or Minnesota Living with Heart Failure Questionnaire), it does
        “refer to meaningful survival a state in which HRQOL [health-related quality of life] is
        satisfactory to the patient.” The guideline also includes quality of life assessments in
        its description of a detailed plan of care for patients with chronic heart failure.
        """
        pass



import arrow

from canvas_workflow_kit import settings

from canvas_workflow_kit.builtin_cqms.cms130v6_colorectal_cancer_screening import ClinicalQualityMeasure130v6
from canvas_workflow_kit.patient_recordset import (
    BillingLineItemRecordSet,
    ConditionRecordSet,
    ImagingReportRecordSet,
    LabReportRecordSet,
    ProtocolOverrideRecordSet,
    ReferralReportRecordSet
)
from canvas_workflow_kit.tests.base import SDKBaseTest
from canvas_workflow_kit.timeframe import Timeframe


class ClinicalQualityMeasure130v6Test(SDKBaseTest):

    def setUp(self):
        self.mocks_path = f'{settings.BASE_DIR}/builtin_cqms/tests/mock_data'

    def test_description(self):
        expected = ('Adults 50-75 years of age who have not had appropriate '
                    'screening for colorectal cancer.')
        self.assertEqual(expected, ClinicalQualityMeasure130v6._meta.description)

    def test_information(self):
        expected = 'https://ecqi.healthit.gov/sites/default/files/ecqm/measures/CMS130v6.html'
        self.assertEqual(expected, ClinicalQualityMeasure130v6._meta.information)

    def test_enabled(self):
        self.assertTrue(ClinicalQualityMeasure130v6.enabled())

    def test_change_types(self):
        result = ClinicalQualityMeasure130v6._meta.compute_on_change_types
        expected = [
            'protocol_override',
            'condition',
            'imaging_report',
            'lab_report',
            'patient',
            'referral_report',
        ]
        self.assertEqual(len(expected), len(result))
        for change in expected:
            self.assertIn(change, result)

    def test_first_due_in(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient older than 50 --> none
        patient = self.load_patient('cms130v6_patient_due')
        patient.patient['birthDate'] = '1968-08-22'
        protocol = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertIsNone(protocol.first_due_in())

        # patient younger than 50 --> 6
        patient = self.load_patient('cms130v6_patient_due')
        patient.patient['birthDate'] = '1968-08-30'
        protocol = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertEqual(6, protocol.first_due_in())

        # patient younger than 50 + colon exclusion --> none
        patient = self.load_patient('cms130v6_patient_due')
        patient.patient['birthDate'] = '1968-08-30'
        patient.conditions = ConditionRecordSet([{
            'coding': [{
                'system': 'http://snomed.info/sct',
                'code': '26390003'
            }],
            'clinicalStatus': 'active',
            'periods': [{
                'from': '2008-08-22',
                'to': None
            }]
        }])
        protocol = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertIsNone(protocol.first_due_in())

    def test_had_colon_exclusion(self):
        patient = self.load_patient('cms130v6_patient_due')
        patient.conditions = ConditionRecordSet([{
            'coding': [{
                'system': 'http://snomed.info/sct',
                'code': '26390003'
            }],
            'clinicalStatus': 'active',
            'periods': [{
                'from': '2008-08-22',
                'to': None
            }]
        }])

        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        protocol = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertTrue(protocol.had_colon_exclusion())

        start = arrow.get('2007-08-23 13:24:56')
        end = arrow.get('2008-08-20 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        protocol = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertFalse(protocol.had_colon_exclusion())

    def test_in_initial_population(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient is 49 --> false
        patient = self.load_patient('cms130v6_patient_due')
        patient.patient['birthDate'] = '1968-08-24'
        protocol = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertFalse(protocol.in_initial_population())

        # patient is 50 --> true
        patient = self.load_patient('cms130v6_patient_due')
        patient.patient['birthDate'] = '1968-08-22'
        protocol = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertTrue(protocol.in_initial_population())

        # patient is 75 --> true
        patient = self.load_patient('cms130v6_patient_due')
        patient.patient['birthDate'] = '1943-08-24'
        protocol = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertTrue(protocol.in_initial_population())

        # patient is 76 --> false
        patient = self.load_patient('cms130v6_patient_due')
        patient.patient['birthDate'] = '1943-08-22'
        protocol = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertFalse(protocol.in_initial_population())

        # context
        patient = self.load_patient('cms130v6_patient_due')
        patient.patient['birthDate'] = '1968-08-22'
        patient.billing_line_items = BillingLineItemRecordSet([])
        tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_initial_population())
        tested = ClinicalQualityMeasure130v6(
            patient=patient, timeframe=timeframe, context='report')
        self.assertFalse(tested.in_initial_population())
        tested = ClinicalQualityMeasure130v6(
            patient=patient, timeframe=timeframe, context='guidance')
        self.assertTrue(tested.in_initial_population())

    def test_in_denominator(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        patient = self.load_patient('cms130v6_patient_due')
        # patient between 50 and 75 --> true
        protocol = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertTrue(protocol.in_denominator())

        # patient with colon exclusion --> false
        patient.conditions = ConditionRecordSet([{
            'coding': [{
                'system': 'http://snomed.info/sct',
                'code': '26390003'
            }],
            'clinicalStatus': 'active',
            'periods': [{
                'from': '2008-08-22',
                'to': None
            }]
        }])
        protocol = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertFalse(protocol.in_denominator())

    def test_in_numerator_with_override(self):
        patient = self.load_patient('cms130v6_patient_satisfied')
        patient.protocol_overrides = ProtocolOverrideRecordSet([{
            'protocolKey': 'ClinicalQualityMeasure130v6',
            'adjustment': {
                'reference': '2018-02-01T00:00:00Z',
                'cycleDays': 180
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable

        #
        patient.lab_reports[0]['originalDate'] = '2018-08-10'
        patient.imaging_reports[0]['originalDate'] = '2018-08-10'
        patient.referral_reports[0]['originalDate'] = '2018-08-10'
        tested = ClinicalQualityMeasure130v6(
            patient=patient, now=arrow.get('2018-08-15').shift(days=0))
        self.assertTrue(tested.in_numerator())
        expected = {
            'date': '2018-08-10',
            'what': 'FOBT',
            'days': 180,
        }
        self.assertEqual(expected, tested._last_exam)
        #
        patient.lab_reports[0]['originalDate'] = '2018-02-10'
        tested = ClinicalQualityMeasure130v6(
            patient=patient, now=arrow.get('2018-08-15').shift(days=0))
        self.assertTrue(tested.in_numerator())
        expected = {
            'date': '2018-08-10',
            'what': 'CT Colonography',
            'days': 180,
        }
        self.assertEqual(expected, tested._last_exam)
        #
        patient.imaging_reports[0]['originalDate'] = '2018-02-10'
        tested = ClinicalQualityMeasure130v6(
            patient=patient, now=arrow.get('2018-08-15').shift(days=0))
        self.assertTrue(tested.in_numerator())
        expected = {
            'date': '2018-08-10',
            'what': 'Colonoscopy',
            'days': 180,
        }
        self.assertEqual(expected, tested._last_exam)
        #
        patient.referral_reports[0]['originalDate'] = '2018-02-10'
        tested = ClinicalQualityMeasure130v6(
            patient=patient, now=arrow.get('2018-08-15').shift(days=0))
        self.assertFalse(tested.in_numerator())

        #
        patient.lab_reports[0]['originalDate'] = '2018-08-10'
        patient.imaging_reports[0]['originalDate'] = '2018-08-10'
        patient.referral_reports[0]['originalDate'] = '2018-08-10'
        tested = ClinicalQualityMeasure130v6(
            patient=patient, now=arrow.get('2018-08-15').shift(days=174))
        self.assertTrue(tested.in_numerator())
        expected = {
            'date': '2018-08-10',
            'what': 'FOBT',
            'days': 180,
        }
        self.assertEqual(expected, tested._last_exam)
        tested = ClinicalQualityMeasure130v6(
            patient=patient, now=arrow.get('2018-08-15').shift(days=176))
        self.assertFalse(tested.in_numerator())

    def test_in_numerator_fitdna(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # referral report
        patient = self.load_patient('cms130v6_patient_due')
        tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())
        self.assertIsNone(tested._last_exam)

        # lab report less than 1 year ago -- > true
        patient = self.load_patient('cms130v6_patient_satisfied')
        patient.lab_reports = LabReportRecordSet([{
            'originalDate': '2018-08-05',
            'loincCodes': [{
                'code': '27925-7'
            }]
        }])
        tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())
        expected = {
            'date': '2018-08-05',
            'what': 'FOBT',
            'days': 365,
        }
        self.assertEqual(expected, tested._last_exam)

        # lab report more than 1 year ago --> false
        patient = self.load_patient('cms130v6_patient_due')
        patient.lab_reports = LabReportRecordSet([{
            'originalDate': '2017-08-05',
            'loincCodes': [{
                'code': '27925-7'
            }]
        }])
        tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())
        self.assertIsNone(tested._last_exam)

        # lab report less than 3 years ago --> true
        patient = self.load_patient('cms130v6_patient_due')
        patient.lab_reports = LabReportRecordSet([{
            'originalDate': '2015-08-26',
            'loincCodes': [{
                'code': '77353-1',
            }]
        }])
        tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())
        expected = {
            'date': '2015-08-26',
            'what': 'FIT-DNA',
            'days': 1096,
        }
        self.assertEqual(expected, tested._last_exam)

        # lab report more than 3 years ago --> false
        patient = self.load_patient('cms130v6_patient_due')
        patient.lab_reports = LabReportRecordSet([{
            'originalDate': '2015-08-17',
            'loincCodes': [{
                'code': '77353-1',
            }]
        }])
        tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())
        self.assertIsNone(tested._last_exam)

    def test_in_numerator_flexible_sigmoidoscopy(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # referral report less than 5 years ago --> true
        patient = self.load_patient('cms130v6_patient_due')
        patient.referral_reports = ReferralReportRecordSet([{
            'originalDate': '2013-08-26',
            'codings': [{
                'system': 'http://www.ama-assn.org/go/cpt',
                'code': '45332',
            }]
        }])
        tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())
        expected = {
            'date': '2013-08-26',
            'what': 'Flexible sigmoidoscopy',
            'days': 1826,
        }
        self.assertEqual(expected, tested._last_exam)

        # referral report more than 5 years ago --> false
        patient = self.load_patient('cms130v6_patient_due')
        patient.referral_reports = ReferralReportRecordSet([{
            'originalDate': '2013-08-17',
            'codings': [{
                'system': 'http://www.ama-assn.org/go/cpt',
                'code': '45332',
            }]
        }])
        tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())
        self.assertIsNone(tested._last_exam)

        # imaging report less than 5 years ago --> true
        patient = self.load_patient('cms130v6_patient_due')
        patient.imaging_reports = ImagingReportRecordSet([{
            'originalDate': '2013-08-26',
            'codings': [{
                'system': 'http://www.ama-assn.org/go/cpt',
                'code': '45332',
            }]
        }])
        tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())
        expected = {
            'date': '2013-08-26',
            'what': 'Flexible sigmoidoscopy',
            'days': 1826,
        }
        self.assertEqual(expected, tested._last_exam)

        # imaging report more than 5 years ago --> false
        patient = self.load_patient('cms130v6_patient_due')
        patient.imaging_reports = ImagingReportRecordSet([{
            'originalDate': '2013-08-17',
            'codings': [{
                'system': 'http://www.ama-assn.org/go/cpt',
                'code': '45332',
            }]
        }])
        tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())
        self.assertIsNone(tested._last_exam)

    def test_in_numerator_ct_colonography(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        tests = [
            ('http://www.ama-assn.org/go/cpt', '74263'),
            ('http://loinc.org', '79101-2'),
        ]
        for system, code in tests:
            # imaging report less than 5 years ago --> true
            patient = self.load_patient('cms130v6_patient_due')
            patient.imaging_reports = ImagingReportRecordSet([{
                'originalDate': '2013-08-26',
                'codings': [{
                    'system': system,
                    'code': code,
                }]
            }])
            tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
            self.assertTrue(tested.in_numerator())
            expected = {
                'date': '2013-08-26',
                'what': 'CT Colonography',
                'days': 1826,
            }
            self.assertEqual(expected, tested._last_exam)

            # imaging report more than 5 years ago --> false
            patient = self.load_patient('cms130v6_patient_due')
            patient.imaging_reports = ImagingReportRecordSet([{
                'originalDate': '2013-08-20',
                'codings': [{
                    'system': system,
                    'code': code,
                }]
            }])
            tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
            self.assertFalse(tested.in_numerator())
            self.assertIsNone(tested._last_exam)

            # referral report less than 5 years ago --> true
            patient = self.load_patient('cms130v6_patient_due')
            patient.referral_reports = ReferralReportRecordSet([{
                'originalDate': '2013-08-26',
                'codings': [{
                    'system': system,
                    'code': code,
                }]
            }])
            tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
            self.assertTrue(tested.in_numerator())
            expected = {
                'date': '2013-08-26',
                'what': 'CT Colonography',
                'days': 1826,
            }
            self.assertEqual(expected, tested._last_exam)

            # referral report more than 5 years ago --> false
            patient = self.load_patient('cms130v6_patient_due')
            patient.referral_reports = ReferralReportRecordSet([{
                'originalDate': '2013-08-20',
                'codings': [{
                    'system': system,
                    'code': code,
                }]
            }])
            tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
            self.assertFalse(tested.in_numerator())
            self.assertIsNone(tested._last_exam)

    def test_in_numerator_colonoscopy(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # referral report less than 10 years ago --> true
        patient = self.load_patient('cms130v6_patient_due')
        patient.referral_reports = ReferralReportRecordSet([{
            'originalDate': '2008-08-26',
            'codings': [{
                'system': 'http://snomed.info/sct',
                'code': '34264006',
            }]
        }])
        tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())
        expected = {
            'date': '2008-08-26',
            'what': 'Colonoscopy',
            'days': 3652,
        }
        self.assertEqual(expected, tested._last_exam)

        # referral report more than 10 years ago --> false
        patient = self.load_patient('cms130v6_patient_due')
        patient.referral_reports = ReferralReportRecordSet([{
            'originalDate': '2008-08-17',
            'codings': [{
                'system': 'http://snomed.info/sct',
                'code': '34264006',
            }]
        }])
        tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())
        self.assertIsNone(tested._last_exam)

        # imaging report less than 10 years ago --> true
        patient = self.load_patient('cms130v6_patient_due')
        patient.imaging_reports = ImagingReportRecordSet([{
            'originalDate': '2008-08-26',
            'codings': [{
                'system': 'http://snomed.info/sct',
                'code': '34264006',
            }]
        }])
        tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())
        expected = {
            'date': '2008-08-26',
            'what': 'Colonoscopy',
            'days': 3652,
        }
        self.assertEqual(expected, tested._last_exam)

        # imaging report more than 10 years ago --> false
        patient = self.load_patient('cms130v6_patient_due')
        patient.imaging_reports = ImagingReportRecordSet([{
            'originalDate': '2008-08-17',
            'codings': [{
                'system': 'http://snomed.info/sct',
                'code': '34264006',
            }]
        }])
        tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())
        self.assertIsNone(tested._last_exam)

    def test_compute_results(self):
        start = arrow.get('2017-10-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        now = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient satisfied
        patient = self.load_patient('cms130v6_patient_satisfied')
        tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        text = 'Nat had a FOBT 2 weeks ago on 8/5/18.'
        self.assertEqual(text, result.narrative)
        self.assertEqual([], result.recommendations)
        self.assertEqual(285, result.due_in)  # not one year since the time frame define the period

        # patient due
        patient = self.load_patient('cms130v6_patient_due')
        tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        text = ('Nat is due for a Colorectal Cancer Screening.\n'
                'No relevant exams found.\n'
                'Current screening interval 10 years.')
        self.assertEqual(text, result.narrative)

        expected = [
            {
                'title': 'Order a FOBT',
                'command': {
                    'type': 'labOrder',
                    'filter': {
                        'coding': [
                            {
                                'system': 'loinc',
                                'code': ['29771-3', '12504-7', '57905-2', '80372-6',
                                         '27396-1', '56491-4', '58453-2', '12503-9',
                                         '2335-8', '14564-9', '27926-5', '27925-7',
                                         '27401-9', '14563-1', '14565-6',
                                         '56490-6']}]}
                },
                'key': 'CMS130v6_RECOMMEND_FOBT',
                'rank': 1,
                'button': 'Order',
            },
            {
                'title': 'Order a FIT-DNA',
                'command': {
                    'type': 'labOrder',
                    'filter': {
                        'coding': [
                            {
                                'system': 'loinc',
                                'code': ['77353-1', '77354-9']}]}
                },
                'key': 'CMS130v6_RECOMMEND_FITDNA',
                'rank': 2,
                'button': 'Order',
            },
            {
                'title': 'Order a Flexible sigmoidoscopy',
                'command': {
                    'type': 'refer',
                    'filter': {
                        'coding': [
                            {
                                'system': 'cpt',
                                'code': ['45350', '45338', '45334', '45337', '45339',
                                         '45333', '45341', '45345', '45346', '45340',
                                         '45349', '45335', '45347', '45330', '45342',
                                         '45332', '45331',
                                         ]},
                            {
                                'system': 'hcpcs', 'code': ['G0104', ]},
                            {
                                'system': 'snomedct',
                                'code': ['396226005', '425634007', '44441009',
                                         '112870002',
                                         ]}]}
                },
                'key': 'CMS130v6_RECOMMEND_SIGMOIDOSCOPY',
                'rank': 3,
                'button': 'Order',
            },
            {
                'title': 'Order a CT Colonography',
                'command': {
                    'type': 'imagingOrder',
                    'filter': {
                        'coding': [
                            {
                                'system': 'cpt',
                                'code': ['74263']},
                            {
                                'system': 'snomedct',
                                'code': ['418714002']}]}
                },
                'key': 'CMS130v6_RECOMMEND_COLONOGRAPHY',
                'rank': 4,
                'button': 'Order',
            },
            {
                'title': 'Order a Colonoscopy',
                'command': {
                    'type': 'refer',
                    'filter': {
                        'coding': [
                            {
                                'system': 'cpt',
                                'code': ['44401', '44402', '45391', '45392', '45388',
                                         '45393', '44394', '44407', '44405', '45387',
                                         '45355', '45398', '45378', '44404', '44392',
                                         '44390', '44403', '44391', '45383', '45384',
                                         '45390', '45379', '45380', '44388', '44393',
                                         '45385', '44397', '45381', '45382', '44406',
                                         '45389', '44389', '44408', '45386',
                                         ]},
                            {
                                'system': 'hcpcs',
                                'code': ['G0105', 'G0121', ]},
                            {
                                'system': 'snomedct',
                                'code': ['8180007', '427459009', '310634005',
                                         '174184006', '443998000', '446521004',
                                         '303587008', '12350003', '367535003',
                                         '444783004', '34264006', '447021001',
                                         '73761001', '25732003', '446745002',
                                         '235150006', '235151005', '174158000',
                                         ]}]}
                },
                'key': 'CMS130v6_RECOMMEND_COLONOSCOPY',
                'rank': 5,
                'button': 'Order',
            }
        ]  # yapf: disable
        # self.assertEqual(recommendations, result.recommendations)
        self.helper_compare_recommendations(expected, result.recommendations)
        self.assertEqual(-1, result.due_in)

        # patient younger than 50 --> n/a + due in
        patient = self.load_patient('cms130v6_patient_due')
        patient.patient['birthDate'] = '1968-08-30'
        tested = ClinicalQualityMeasure130v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertEqual(6, result.due_in)

    def test_compute_results_with_override(self):
        patient = self.load_patient('cms130v6_patient_satisfied')
        patient.protocol_overrides = ProtocolOverrideRecordSet([{
            'protocolKey': 'ClinicalQualityMeasure130v6',
            'adjustment': {
                'cycleDays': 180
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable
        patient.referral_reports[0]['originalDate'] = '2019-07-26'
        patient.billing_line_items[0]['created'] = '2019-07-26'
        patient.billing_line_items[0]['datetimeOfService'] = '2019-07-26'

        # exam done within the last 180 days
        tested = ClinicalQualityMeasure130v6(
            patient=patient, now=arrow.get('2019-07-26').shift(days=80))
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        text = 'Nat had a Colonoscopy 3 months ago on 7/26/19.'
        self.assertEqual(text, result.narrative)
        self.assertEqual(0, len(result.recommendations))
        self.assertEqual(100, result.due_in)

        # exam done 181 days
        patient.billing_line_items[0]['created'] = '2028-07-29'  # visit within the 180-day period
        patient.billing_line_items[0]['datetimeOfService'] = '2028-07-29'  # visit within the 180-day period
        tested = ClinicalQualityMeasure130v6(
            patient=patient, now=arrow.get('2028-07-26').shift(days=181))
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        text = ('Nat is due for a Colorectal Cancer Screening.\n'
                'Last Colonoscopy done 9 years ago on 7/26/19.\n'
                'Current screening interval 6 months.')
        self.assertEqual(text, result.narrative)
        self.assertEqual(5, len(result.recommendations))
        self.assertEqual(-1, result.due_in)

    def helper_compare_recommendations(self, expected, result):
        for exp, recommendation in zip(expected, result):
            self.assertEqual(exp['key'], recommendation.key)
            self.assertEqual(exp['rank'], recommendation.rank)
            self.assertEqual(exp['button'], recommendation.button)
            self.assertEqual(exp['title'], recommendation.title)
            self.assertEqual(exp['command']['type'], recommendation.command['type'])
            exp_item = exp['command']['filter']['coding']
            rec_item = recommendation.command['filter']['coding']
            self.assertEqual(len(exp_item), len(rec_item))
            for exp_coding, rec_coding in zip(exp_item, rec_item):
                self.assertEqual(exp_coding['system'], rec_coding['system'])
                self.assertEqual(sorted(exp_coding['code']), sorted(rec_coding['code']))



from canvas_workflow_kit import events
from canvas_workflow_kit.protocol import (
    STATUS_DUE,
    STATUS_SATISFIED,
    ClinicalQualityMeasure,
    ProtocolResult
)
from canvas_workflow_kit.recommendation import Recommendation
from canvas_workflow_kit.value_set.medication_class_path2018 import Antiarrhythmics
from canvas_workflow_kit.value_set.specials import DysrhythmiaClassConditionSuspect


class Hcc004v1(ClinicalQualityMeasure):

    class Meta:
        title = 'Dysrhythmia Suspects'
        version = '2019-02-12v1'
        description = ('All patients with potential dysrhythmia based on an '
                       'active medication without associated active problem.')
        information = 'https://canvas-medical.zendesk.com/hc/en-us/articles/360059083773-Dysrhythmia-Suspects-HCC004v1'  # noqa: E501

        identifiers = ['HCC004v1']

        types = ['HCC']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]
        authors = [
            'Canvas Medical Team',
        ]

        references = [
            'Canvas Medical HCC, https://canvas-medical.zendesk.com/hc/en-us/articles/360059083773-Dysrhythmia-Suspects-HCC004v1'  # noqa: E501
        ]

        compute_on_change_types = [
            ClinicalQualityMeasure.CHANGE_PROTOCOL_OVERRIDE,
            ClinicalQualityMeasure.CHANGE_CONDITION,
            ClinicalQualityMeasure.CHANGE_MEDICATION,
        ]

    def in_initial_population(self) -> bool:
        return True

    def in_denominator(self) -> bool:
        """
        Patients with any active medication in Antiarrhythmics Drug Class
        """
        if self.patient.medications.find(Antiarrhythmics).intersects(
                self.timeframe, still_active=self.patient.active_only):
            return True
        return False

    def in_numerator(self) -> bool:
        """
        Patients without active conditions within the list with ICD 10 I42.* I47.*, I48.*, I49.*
        """
        if (self.patient.conditions
                        .intersects(self.timeframe, still_active=self.patient.active_only)
                        .find_class(DysrhythmiaClassConditionSuspect)):  # yapf: disable
            return False
        return True

    def compute_results(self) -> ProtocolResult:
        result = ProtocolResult()
        if self.in_denominator():
            if self.in_numerator():
                result.due_in = -1
                result.status = STATUS_DUE
                result.add_narrative((
                    '{0} has an active medication on the Medication List commonly used for Dysrhythmia. '  # noqa: E501
                    'There is no associated condition on the Conditions List.').format(
                        self.patient.first_name))
                title = ('Consider updating the Conditions List to include Dysrhythmia '
                         'related problem as clinically appropriate.')
                result.add_recommendation(
                    Recommendation(
                        key='HCC004v1_RECOMMEND_DIAGNOSE_DYSRHYTHMIA',
                        rank=1,
                        button='Diagnose',
                        title=title,
                        narrative=result.narrative,
                        command={'key': 'diagnose'}))
            else:
                result.status = STATUS_SATISFIED
        return result



from xml.dom.minidom import parseString as parse_xml

from pygments import highlight
from pygments.lexers.html import XmlLexer
from pygments.formatters.terminal256 import TerminalTrueColorFormatter

from pyxb import ValidationError
from pyxb.utils.domutils import BindingDOMSupport

from schemas import eMeasure


def prettify(xml):
    return parse_xml(xml).toprettyxml(indent='  ', encoding='utf-8')


class QualityMeasureDocument(object):

    @classmethod
    def from_file(cls, path):
        qmd = cls()

        qmd.xml = open(path).read()
        qmd.document = eMeasure.CreateFromDocument(qmd.xml)

        return qmd

    @staticmethod
    def node_to_xml(node):
        try:
            dom = node.toDOM()

            # Specify the namespace of the binding class as the default for more compact XML (and
            # because this is how it appears in the example XML from Surescripts)
            dom = node.toDOM(bds=BindingDOMSupport(default_namespace=eMeasure.Namespace))

            return prettify(dom.toxml('utf-8'))
        except ValidationError as e:
            print(e.details())

            raise

    def to_xml(self):
        return self.node_to_xml(self.document)

    def print_node(self, node):
        self.print_xml(self.node_to_xml(node).decode())

    def print_document(self):
        self.print_xml(self.to_xml().decode())

    @staticmethod
    def print_xml(xml):
        highlighted = highlight(xml, XmlLexer(), TerminalTrueColorFormatter(style='paraiso-dark'))
        decorated = '\n'.join(line for line in highlighted.strip().split('\n'))

        print(decorated)
        print()



from ..value_set import ValueSet


class PositiveFinding(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent positive test results. This is intended to be paired with other concepts that identify specific medical tests.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) attribute related to Result.

    **Inclusion Criteria:** Includes only relevant concepts associated with identifying a positive test result.

    **Exclusion Criteria:** Excludes concepts that identify a specific type of medical test.
    """

    OID = '2.16.840.1.113883.3.464.1003.121.12.1016'
    VALUE_SET_NAME = 'Positive Finding'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    SNOMEDCT = {
        '441773004'
    }

class TobaccoNonUser(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent tobacco non-user status.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) attribute related to Result.

    **Inclusion Criteria:** Includes only relevant concepts associated with indicating a patient does not use tobacco products, including smoking and smoke-less tobacco products such as chew, snuff, pipe, cigarette, cigar, etc.

    **Exclusion Criteria:** Excludes concepts that may indicate a current tobacco user status.
    """

    OID = '2.16.840.1.113883.3.526.3.1189'
    VALUE_SET_NAME = 'Tobacco Non-User'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    SNOMEDCT = {
        '105539002',
        '105540000',
        '105541001',
        '160618006',
        '160620009',
        '160621008',
        '228491005',
        '228492003',
        '228493008',
        '228501004',
        '228502006',
        '228503001',
        '228511006',
        '228512004',
        '228513009',
        '266919005',
        '266921000',
        '266922007',
        '266923002',
        '266924008',
        '266925009',
        '266928006',
        '281018007',
        '360890004',
        '360900008',
        '360918006',
        '360929005',
        '405746006',
        '428081000124100',
        '428091000124102',
        '451371000124109',
        '451381000124107',
        '456711000124105',
        '48031000119106',
        '53896009',
        '702975009',
        '702979003',
        '735128000',
        '8392000',
        '8517006',
        '87739003'
    }

class VisualAcuity2040OrBetter(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent visual acuity findings that are 20/40 or better.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) attribute related to Result.

    **Inclusion Criteria:** Includes only relevant concepts associated with distance vision findings of 20/40 or better.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.526.3.1483'
    VALUE_SET_NAME = 'Visual Acuity 20/40 or Better'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    SNOMEDCT = {
        '422497000',
        '423059004',
        '423364005',
        '423862000',
        '424703005'
    }


from typing import Dict, List
from unittest.mock import PropertyMock, patch

import arrow

from canvas_workflow_kit import settings

from canvas_workflow_kit.builtin_cqms.ccp001v1_symptomatic_surveillance import Ccp001v1
from canvas_workflow_kit.patient_recordset import InterviewRecordSet
from canvas_workflow_kit.tests.base import SDKBaseTest
from canvas_workflow_kit.timeframe import Timeframe


class TestCcp001v1(SDKBaseTest):

    def setUp(self):
        self.mocks_path = f'{settings.BASE_DIR}/builtin_cqms/tests/mock_data'

    def test_enabled(self):
        self.assertTrue(Ccp001v1.enabled())

    def test_description(self):
        # flake8: noqa
        expected = 'All patients with COVID Questionnaire completed Date < 7 days ago and >  5 days ago.'
        self.assertEqual(expected, Ccp001v1._meta.description)

    def test_information(self):
        # flake8: noqa
        expected = 'https://canvas-medical.zendesk.com/hc/en-us/articles/360059084173-COVID-19-Risk-Assessment-Follow-Up-Protocol'
        self.assertEqual(expected, Ccp001v1._meta.information)

    def test_change_types(self):
        result = Ccp001v1._meta.compute_on_change_types
        expected = ['interview']
        self.assertEqual(expected, result)

    def test___init__(self):
        start = arrow.get('2019-03-20 13:24:56')
        end = arrow.get('2020-03-20 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('patient')
        tested = Ccp001v1(patient=patient, timeframe=timeframe)

        self.assertEqual([], tested.rationales_for_inclusion_in_numerator)

    def test_anchor_time(self):
        patient = self.load_patient('patient')

        tests = [
            ('00', '2020-03-20T09:00:00+00:00'),
            ('01', '2020-03-20T09:00:00+00:00'),
            ('02', '2020-03-20T09:00:00+00:00'),
            ('03', '2020-03-20T09:00:00+00:00'),
            ('04', '2020-03-20T09:00:00+00:00'),
            ('05', '2020-03-20T09:00:00+00:00'),
            ('06', '2020-03-20T09:00:00+00:00'),
            ('07', '2020-03-20T09:00:00+00:00'),
            ('08', '2020-03-20T09:00:00+00:00'),
            ('09', '2020-03-20T09:00:00+00:00'),
            ('10', '2020-03-20T09:00:00+00:00'),
            ('11', '2020-03-20T09:00:00+00:00'),
            ('12', '2020-03-20T09:00:00+00:00'),
            ('13', '2020-03-20T09:00:00+00:00'),
            ('14', '2020-03-20T09:00:00+00:00'),
            ('15', '2020-03-20T09:00:00+00:00'),
            ('16', '2020-03-20T09:00:00+00:00'),
            ('17', '2020-03-20T09:00:00+00:00'),
            ('18', '2020-03-20T09:00:00+00:00'),
            ('19', '2020-03-20T09:00:00+00:00'),
            ('20', '2020-03-20T09:00:00+00:00'),
            ('21', '2020-03-21T09:00:00+00:00'),
            ('22', '2020-03-21T09:00:00+00:00'),
            ('23', '2020-03-21T09:00:00+00:00'),
        ]
        start = arrow.get('2019-03-20 13:24:56')
        for hour, expected in tests:
            end = arrow.get(f'2020-03-20 {hour}:24:56')
            timeframe = Timeframe(start=start, end=end)
            tested = Ccp001v1(patient=patient, timeframe=timeframe)
            result = tested.anchor_time
            assert expected == str(result), f'{hour} -> {result}'

    @patch(
        'canvas_workflow_kit.builtin_cqms.ccp001v1_symptomatic_surveillance.Ccp001v1.anchor_time',
        new_callable=PropertyMock)
    def test_interview(self, anchor_time):
        anchor_time.return_value = arrow.get('2020-03-20T09:00:00+00:00')
        start = arrow.get('2019-03-20 13:24:56')
        end = arrow.get('2010-03-20 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('patient')

        # no interview -> empty
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        result = tested.interview
        self.assertEqual(0, len(result.keys()))

        # interview just after the end of the 6th day -> empty
        patient.interviews = InterviewRecordSet(
            [self.helper_questionnaire(123, '2020-03-14T08:59:59+00:00', {})])
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        result = tested.interview
        self.assertEqual(0, len(result.keys()))

        # interview just before the end of the 6th day -> the interview
        patient.interviews = InterviewRecordSet(
            [self.helper_questionnaire(124, '2020-03-14T09:00:00+00:00', {})])
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        result = tested.interview
        self.assertEqual(7, len(result.keys()))
        self.assertEqual(124, result['id'])

        # interview within the 6th day -> the interview
        for day in range(14, 20):
            patient.interviews = InterviewRecordSet(
                [self.helper_questionnaire(124, f'2020-03-{day}T17:45:41+00:00', {})])
            tested = Ccp001v1(patient=patient, timeframe=timeframe)
            result = tested.interview
            self.assertEqual(7, len(result.keys()))
            self.assertEqual(124, result['id'])

        # several interviews -> the interview that is the oldest within the 6 days
        patient.interviews = InterviewRecordSet([
            self.helper_questionnaire(124, '2020-03-14T08:59:59+00:00', {}),
            self.helper_questionnaire(144, '2020-03-14T09:00:10+00:00', {}),
            self.helper_questionnaire(134, '2020-03-14T09:00:09+00:00', {}),
            self.helper_questionnaire(154, '2020-03-14T09:00:11+00:00', {}),
        ])
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        result = tested.interview
        self.assertEqual(7, len(result.keys()))
        self.assertEqual(134, result['id'])

    @patch(
        'canvas_workflow_kit.builtin_cqms.ccp001v1_symptomatic_surveillance.Ccp001v1.anchor_time',
        new_callable=PropertyMock)
    def test_in_initial_population(self, anchor_time):
        anchor_time.return_value = arrow.get('2020-03-20T09:00:00+00:00')
        start = arrow.get('2019-03-20 13:24:56')
        end = arrow.get('2010-03-20 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('patient')

        # no interview -> no
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_initial_population())

        # interview older than 6 days -> no
        patient.interviews = InterviewRecordSet(
            [self.helper_questionnaire(123, '2020-03-14T08:59:59+00:00', {})])
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_initial_population())

        # interview newer than 6 days -> yes
        for day in range(14, 20):
            patient.interviews = InterviewRecordSet(
                [self.helper_questionnaire(124, f'2020-03-{day}T17:45:41+00:00', {})])
            tested = Ccp001v1(patient=patient, timeframe=timeframe)
            self.assertTrue(tested.in_initial_population())

        # several interviews with some within 6 days -> yes
        patient.interviews = InterviewRecordSet([
            self.helper_questionnaire(124, '2020-03-14T08:59:59+00:00', {}),
            self.helper_questionnaire(144, '2020-03-14T09:00:10+00:00', {}),
            self.helper_questionnaire(134, '2020-03-14T09:00:09+00:00', {}),
            self.helper_questionnaire(154, '2020-03-14T09:00:11+00:00', {}),
        ])
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_initial_population())

    @patch(
        'canvas_workflow_kit.builtin_cqms.ccp001v1_symptomatic_surveillance.Ccp001v1.anchor_time',
        new_callable=PropertyMock)
    def test_in_denominator(self, anchor_time):
        anchor_time.return_value = arrow.get('2020-03-20T09:00:00+00:00')
        start = arrow.get('2019-03-20 13:24:56')
        end = arrow.get('2010-03-20 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('patient')

        # no interview -> no
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

        # interviews
        tests = [
            ('2020-03-14T08:59:59+00:00', False),  # interview older than 6 days -> no
            ('2020-03-14T09:00:00+00:00', True),  # interview within 6 days -> yes
            ('2020-03-14T22:33:44+00:00', True),  # interview within 6 days -> yes
            ('2020-03-15T08:59:59+00:00', True),  # interview within 6 days -> yes
            ('2020-03-15T09:00:00+00:00', False),  # interview newer than 6 days -> no
        ]
        for date, expected in tests:
            patient.interviews = InterviewRecordSet([self.helper_questionnaire(123, date, {})])
            tested = Ccp001v1(patient=patient, timeframe=timeframe)
            if expected:
                self.assertTrue(tested.in_denominator())
            else:
                self.assertFalse(tested.in_denominator())

    def test_group_question_responses(self):
        tested = Ccp001v1.group_question_responses

        # empty --> empty
        interview = {
            'questions': [],
            'responses': [],
        }
        result = tested(interview)
        self.assertEqual({}, result)

        #
        interview = {
            'questions': [
                {
                    'questionResponseId': 10,
                    'code': 'question-10'
                },
                {
                    'questionResponseId': 11,
                    'code': 'question-11'
                },
                {
                    'questionResponseId': 12,
                    'code': 'question-12'
                },
            ],
            'responses': [
                {
                    'questionResponseId': 10,
                    'code': 'response-10'
                },
                {
                    'questionResponseId': 11,
                    'code': 'response-11a'
                },
                {
                    'questionResponseId': 11,
                    'code': 'response-11b'
                },
                {
                    'questionResponseId': 11,
                    'code': 'response-11c'
                },
                {
                    'questionResponseId': 12,
                    'code': 'response-12'
                },
            ],
        }
        result = tested(interview)
        expected = {
            'question-10': ['response-10'],
            'question-11': ['response-11a', 'response-11b', 'response-11c'],
            'question-12': ['response-12'],
        }
        self.assertEqual(expected, result)

    @patch(
        'canvas_workflow_kit.builtin_cqms.ccp001v1_symptomatic_surveillance.Ccp001v1.interview',
        new_callable=PropertyMock)
    @patch(
        'canvas_workflow_kit.builtin_cqms.ccp001v1_symptomatic_surveillance.Ccp001v1.anchor_time',
        new_callable=PropertyMock)
    def test_in_numerator(self, anchor_time, interview):
        anchor_time.return_value = arrow.get('2020-03-20T09:00:00+00:00')
        start = arrow.get('2019-03-20 13:24:56')
        end = arrow.get('2010-03-20 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('patient')

        # all responses are negative --> false
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        interview.return_value = self.helper_questionnaire(123, '2020-03-16 20:24:56+00:00', {})
        self.assertFalse(tested.in_numerator())
        self.assertEqual([], tested.rationales_for_inclusion_in_numerator)

        # Q1 has one answer -> true + value in narrative
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        interview.return_value = self.helper_questionnaire(123, '2020-03-16 20:24:56+00:00',
                                                           {35: '386661006'})
        self.assertTrue(tested.in_numerator())
        expected = ['Symptoms exhibited: fever']
        self.assertEqual(expected, tested.rationales_for_inclusion_in_numerator)

        # Q1 has two answer -> true + all values in narrative
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        interview.return_value = self.helper_questionnaire(123, '2020-03-16 20:24:56+00:00', {
            35: '386661006',
            39: '49727002'
        })
        self.assertTrue(tested.in_numerator())
        expected = ['Symptoms exhibited: fever, cough']
        self.assertEqual(expected, tested.rationales_for_inclusion_in_numerator)

        # Q4 yes -> true + value in narrative
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        interview.return_value = self.helper_questionnaire(123, '2020-03-16 20:24:56+00:00',
                                                           {38: '183616001'})
        self.assertTrue(tested.in_numerator())
        expected = ['Follow-up requested by care team']
        self.assertEqual(expected, tested.rationales_for_inclusion_in_numerator)

        # Q2 yes + patient 65+ -> yes
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        interview.return_value = self.helper_questionnaire(123, '2020-03-16 20:24:56+00:00',
                                                           {36: '276030007'})
        self.assertTrue(tested.in_numerator())
        expected = ['Age 65+ with recent travel']
        self.assertEqual(expected, tested.rationales_for_inclusion_in_numerator)

        # Q2 yes + patient < 65> -> no
        patient.patient['birthDate'] = '1955-03-22'
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        interview.return_value = self.helper_questionnaire(123, '2020-03-16 20:24:56+00:00',
                                                           {36: '276030007'})
        self.assertFalse(tested.in_numerator())
        self.assertEqual([], tested.rationales_for_inclusion_in_numerator)
        patient.patient['birthDate'] = '1954-03-22'

        # Q3 yes + patient 65+ -> yes
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        interview.return_value = self.helper_questionnaire(123, '2020-03-16 20:24:56+00:00',
                                                           {37: '150781000119103'})
        self.assertTrue(tested.in_numerator())
        expected = ['Age 65+ with recent exposure']
        self.assertEqual(expected, tested.rationales_for_inclusion_in_numerator)

        # Q3 yes + patient < 65> -> no
        patient.patient['birthDate'] = '1955-03-22'
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        interview.return_value = self.helper_questionnaire(123, '2020-03-16 20:24:56+00:00',
                                                           {37: '150781000119103'})
        self.assertFalse(tested.in_numerator())
        self.assertEqual([], tested.rationales_for_inclusion_in_numerator)
        patient.patient['birthDate'] = '1954-03-22'

        # Q1, Q2, Q3, Q4
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        responses = {
            35: '386661006',
            36: '276030007',
            37: '150781000119103',
            38: '183616001',
            39: '49727002',
        }
        interview.return_value = self.helper_questionnaire(123, '2020-03-16 20:24:56+00:00',
                                                           responses)
        self.assertTrue(tested.in_numerator())
        expected = [
            'Symptoms exhibited: fever, cough',
            'Follow-up requested by care team',
            'Recent travel',
            'Recent exposure',
        ]
        self.assertEqual(expected, tested.rationales_for_inclusion_in_numerator)

    @patch('canvas_workflow_kit.builtin_cqms.ccp001v1_symptomatic_surveillance.arrow.utcnow')
    @patch(
        'canvas_workflow_kit.builtin_cqms.ccp001v1_symptomatic_surveillance.Ccp001v1.anchor_time',
        new_callable=PropertyMock)
    def test_compute_results(self, anchor_time, utcnow):
        utcnow.return_value = arrow.get('2020-03-20 13:24:56')
        anchor_time.return_value = arrow.get('2020-03-20T09:00:00+00:00')

        start = arrow.get('2019-03-20 13:24:56')
        end = arrow.get('2010-03-20 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('patient')

        # no interview -> n/a with no due_in
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertEqual([], result.narratives)
        self.assertEqual([], result.recommendations)
        self.assertIsNone(result.due_in)
        self.assertIsNone(result.next_review)

        # too old interview -> n/a with no due_in
        patient.interviews = InterviewRecordSet(
            [self.helper_questionnaire(123, '2020-03-14T08:59:59+00:00', {})])
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertEqual([], result.narratives)
        self.assertEqual([], result.recommendations)
        self.assertIsNone(result.due_in)
        self.assertIsNone(result.next_review)

        # too new interview -> n/a with due_in=1
        patient.interviews = InterviewRecordSet(
            [self.helper_questionnaire(123, '2020-03-16T09:01:02+00:00', {})])
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertEqual([], result.narratives)
        self.assertEqual([], result.recommendations)
        self.assertEqual(1, result.due_in)
        next_review = arrow.get('2020-03-21T10:00:00+00:00')
        self.assertEqual(next_review, result.next_review)

        # too new interview -> n/a with due_in=3
        patient.interviews = InterviewRecordSet(
            [self.helper_questionnaire(123, '2020-03-19T09:01:02+00:00', {})])
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertEqual([], result.narratives)
        self.assertEqual([], result.recommendations)
        self.assertEqual(4, result.due_in)
        next_review = arrow.get('2020-03-21T10:00:00+00:00')
        self.assertEqual(next_review, result.next_review)

        # the review time is constant regardless of the current time
        # as long as it is in the window
        tests = [
            ('2020-03-19 20:24:56', '2020-03-20T10:00:00+00:00', 5),
            ('2020-03-19 22:24:57', '2020-03-21T10:00:00+00:00', 5),
            ('2020-03-22 22:24:57', '2020-03-24T10:00:00+00:00', 2),
        ]
        for utcnow_time, expected_review, expected_due in tests:
            utcnow.return_value = arrow.get(utcnow_time)
            patient.interviews = InterviewRecordSet(
                [self.helper_questionnaire(123, '2020-03-19T09:01:02+00:00', {})])
            tested = Ccp001v1(patient=patient, timeframe=timeframe)
            result = tested.compute_results()
            self.assertEqual('not_applicable', result.status)
            self.assertEqual([], result.narratives)
            self.assertEqual([], result.recommendations)
            self.assertEqual(expected_due, result.due_in)
            next_review = arrow.get(expected_review)
            self.assertEqual(next_review, result.next_review)

        # interview done 6 days ago without follow up -> satisfied
        utcnow.return_value = arrow.get('2020-03-20 13:24:56')
        patient.interviews = InterviewRecordSet(
            [self.helper_questionnaire(123, '2020-03-14T13:24:58+00:00', {})])
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        self.assertEqual(['Nicolas does not need a follow-up arranged'], result.narratives)
        self.assertEqual([], result.recommendations)
        self.assertEqual(-1, result.due_in)
        next_review = arrow.get('2020-03-21T13:24:56+00:00')
        self.assertEqual(next_review, result.next_review)

        # interview done 6 days ago requiring a follow up -> due
        responses = {
            35: '386661006',
            36: '276030007',
            37: '150781000119103',
            38: '183616001',
            39: '49727002',
        }
        patient.interviews = InterviewRecordSet(
            [self.helper_questionnaire(123, '2020-03-14T20:24:56+00:00', responses)])
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        narratives = [
            'Nicolas should have a follow-up arranged',
            'Symptoms exhibited: fever, cough',
            'Follow-up requested by care team',
            'Recent travel',
            'Recent exposure',
        ]
        self.assertEqual(narratives, result.narratives)
        self.assertEqual(1, len(result.recommendations))
        recommendation = result.recommendations[0]
        self.assertEqual('CCP001v1_RECOMMEND_FOLLOW_UP', recommendation.key)
        self.assertEqual('Arrange a follow-up', recommendation.title)
        self.assertEqual(-1, result.due_in)
        next_review = arrow.get('2020-03-21 13:24:56+00:00')
        self.assertEqual(next_review, result.next_review)

        # interview done 7 days ago -> not applicable
        patient.interviews = InterviewRecordSet(
            [self.helper_questionnaire(123, '2020-03-13T13:24:58+00:00', {})])
        tested = Ccp001v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertEqual([], result.narratives)
        self.assertEqual([], result.recommendations)
        self.assertIsNone(result.due_in)
        self.assertIsNone(result.next_review)

    def helper_questionnaire(self, interview_id: int, date: str, answers: Dict):
        question_codes = {
            35: 'CANVAS0002',
            36: 'CANVAS0003',
            37: 'CANVAS0004',
            38: 'CANVAS0005',
            39: 'CANVAS0002',
        }
        response_values = {
            '386661006': 'fever',
            '267036007': 'sob',
            '49727002': 'cough',
            '276030007': 'travel abroad',
            '150781000119103': 'exposure to viral disease',
            '183616001': 'Follow up arranged',
            '260385009': 'negative',
        }
        default = {
            35: '',
            36: '260385009',
            37: '260385009',
            38: '260385009',
            39: '',
        }
        default.update(answers)

        questions: Dict = {}
        responses: List = []
        for questionResponseId, codes in default.items():
            for code in [c for c in codes.split(',') if c != '']:
                questions[questionResponseId] = {
                    'questionResponseId': questionResponseId,
                    'code': question_codes[questionResponseId],
                    'codeSystem': 'CANVAS',
                }
                responses.append({
                    'questionResponseId': questionResponseId,
                    'value': response_values[code],
                    'code': code,
                    'codeSystem': 'http://snomed.info/sct'
                })

        return {
            'id': interview_id,
            'noteTimestamp': date,
            'name': 'COVID Questionnaire',
            'results': [],
            'questionnaires': [{
                'code': 'CANVAS0001',
                'codeSystem': 'CANVAS'
            }],
            'questions': list(questions.values()),
            'responses': responses,
        }



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    EncounterToDocumentMedications, Ethnicity, OncAdministrativeSex, Payer, Race)


class ClinicalQualityMeasure68v10(ClinicalQualityMeasure):
    """
    Documentation of Current Medications in the Medical Record

    Description: Percentage of visits for patients aged 18 years and older for which the eligible
    professional or eligible clinician attests to documenting a list of current medications using
    all immediate resources available on the date of the encounter

    Definition: Current Medications:
    Medications the patient is presently taking including all prescriptions, over-the-counters,
    herbals and vitamin/mineral/dietary (nutritional) supplements with each medication's name,
    dosage, frequency and administered route.

    Route:
    Documentation of the way the medication enters the body (some examples include but are not
    limited to: oral, sublingual, subcutaneous injections, and/or topical).

    Rationale: According to the National Center for Health Statistics, during the years of
    2011-2014, 46.9 percent of patients (both male and female) were prescribed at least one
    prescription medication with 10.9 percent taking 5 or more medications. Additionally, 90.6
    percent of patients (both male and female) aged 65 years and older were prescribed at least one
    medication with 40.7 percent taking 5 or more medications (2017).  In this context, maintaining
    an accurate and complete medication list has proven to be a challenging documentation endeavor
    for various health care provider settings. While most of outpatient encounters (2/3) result in
    providers prescribing at least one medication, hospitals have been the focus of medication
    safety efforts (Stock, Scott, & Gurtel, 2009). Nassaralla, Naessens, Chaudhry, Hansen, and
    Scheitel (2007) caution that this is at odds with the current trend, where patients with
    chronic illnesses are increasingly being treated in the outpatient setting and require careful
    monitoring of multiple medications. Additionally, Nassaralla et al. (2007) reveal that it is in
    fact in outpatient settings where more fatal adverse drug events (ADE) occur when these are
    compared to those occurring in hospitals (1 of 131 outpatient deaths compared to 1 in 854
    inpatient deaths). In the outpatient setting, ADEs occur 25% of the time and over one-third of
    these are considered preventable (Tache, Sonnichsen, & Ashcroft, 2011). Particularly vulnerable
    are patients over 65 years, with evidence suggesting that the rate of ADEs per 10,000 person
    per year increases with age; 25-44 years old at 1.3; 45-64 at 2.2, and 65 + at 3.8 (Sarkar,
    López, & Maselli, 2011). Another vulnerable group is chronically ill individuals. These
    population groups are more likely to experience ADEs and subsequent hospitalization.

    A multiplicity of providers and inadequate care coordination among them has been identified as
    barriers to collecting complete and reliable medication records. Data indicate that
    reconciliation and documentation continue to be poorly executed with discrepancies occurring in
    92% (74 of 80 patients) of medication lists among admittance to the emergency room. Of 80
    patients included in the study, the home medications were reordered for 65% of patients on
    their admission and of the 65% the majority (29%) had a change in their dosing interval, while
    23% had a change in their route of administration, and 13% had a change in dose. A total of 361
    medication discrepancies, or the difference between the medications patients were taking before
    admission and those listed in their admission orders, were identified in at least 74 patients
    (Poornima et al., 2015). The study found that "Through an appropriate reconciliation programme,
    around 80% of errors relating to medication and the potential harm caused by these errors could
    be reduced" (Penumarthi et al., 2015, p. 243).

    Documentation of current medications in the medical record facilitates the process of
    medication review and reconciliation by the provider, which is necessary for reducing ADEs and
    promoting medication safety. The need for provider to provider coordination regarding
    medication records, and the existing gap in implementation, is highlighted in the American
    Medical Association's Physician's Role in Medication Reconciliation, which states that
    "critical patient information, including medical and medication histories, current medications
    the patient is receiving and taking, and sources of medications, is essential to the delivery
    of safe medical care. However, interruptions in the continuity of care and information gaps in
    patient health records are common and significantly affect patient outcomes" (2007, p. 7). This
    is because clinical decisions based on information that is incomplete and/or inaccurate are
    likely to lead to medication error and ADEs. Weeks, Corbette, and Stream (2010) noted similar
    barriers and identified the utilization of health information technology as an opportunity for
    facilitating the creation of universal medication lists. One 2015 meta-analysis showed an
    association between EHR documentation with an overall RR of 0.46 (95% CI = 0.38 to 0.55; P <
    0.001) and ADEs with an overall RR of 0.66 (95% CI = 0.44 to 0.99; P = 0.045). This meta-
    analysis provides evidence that the use of the EHR can improve the quality of healthcare
    delivered to patients by reducing medication errors and ADEs (Campanella et al., 2016).

    Guidance: This eCQM is an episode-based measure. This measure is to be reported for every
    encounter during the measurement period.

    Eligible professionals or eligible clinicians reporting this measure may document medication
    information received from the patient, authorized representative(s), caregiver(s) or other
    available healthcare resources.

    This list must include all known prescriptions, over-the-counter (OTC) products, herbals,
    vitamins, minerals, dietary (nutritional) supplements AND must contain the medications' name,
    dosage, frequency and route of administration.

    This measure should also be reported if the eligible professional or eligible clinician
    documented the patient is not currently taking any medications.

    By reporting the action described in this measure, the provider attests to having documented a
    list of current medications utilizing all immediate resources available at the time of the
    encounter.

    This version of the eCQM uses QDM version 5.5.  Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms68v10
    """

    title = 'Documentation of Current Medications in the Medical Record'

    identifiers = ['CMS68v10']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'PCPI(R) Foundation (PCPI[R])',
    ]

    references = [
        'American Medical Association. (2007). The physician’s role in medication reconciliation: Issues, strategies, and safety principles. Retrieved from https://pogoe.org/sites/default/files/Medication%20Reconciliation.pdf',
        'Campanella, P., Lovato, E., Marone, C., Fallacara, L., Mancuso, A., Ricciardi, W., & Specchia, M. L. (2016). The impact of electronic health records on health care quality: A systematic review and meta-analysis. European Journal of Public Health, 26(1), 60-64. doi:10.1093/eurpub/ckv122',
        'Nassaralla, C. L., Naessens, J. M., Chaudhry, R., Hansen, M. A., & Scheitel, S. M. (2007). Implementation of a medication reconciliation process in an ambulatory internal medicine clinic. Quality and Safety in Health Care, 16(2), 90-94. doi:10.1136/qshc.2006.021113',
        'National Center for Health Statistics. (2017). Health, United States, 2017: Supplementary Table 79: Prescription drug use in the United States by sex, race, age, and origin. Retrieved from https://www.cdc.gov/nchs/data/hus/2017/079.pdf',
        'National Quality Forum. (2010). Safe practices for better healthcare - 2010 update. Retrieved from http://www.qualityforum.org/Projects/Safe_Practices_2010.aspx',
        'Penumarthi, P., Pasala, R., T V, R., Nagasubramanian, VR., Devi, G S., Seshadri, P. (2015). Medication reconciliation and medication error prevention in an emergency department of a tertiary care hospital. Journal of Young Pharmacists, 7(3), 241-249. doi:10.5530/jyp.2015.3.15',
        'Sarkar, U., López, A., Maselli, J.H., Gonzales, R. (2011). Adverse drug events in U.S. adult ambulatory medical care. Health Services Research, 46(5), 1517-1533. doi:10.1111/j.1475-6773.2011.01269.x',
        'Stock, R., Scott, J., & Gurtel, S. (2009). Using an electronic prescribing system to ensure accurate medication lists in a large multidisciplinary medical group. The Joint Commission Journal on Quality and Patient Safety, 35(5), 271-277.',
        'Tache, S. V., Sonnichsen, A., & Ashcroft, D. M. (2011). Prevalence of adverse drug events in ambulatory care: A systematic review. The Annals of Pharmacotherapy, 45(7-8), 977-989. doi: 10.1345/aph.1P627',
        'The Joint Commission. (2019). Ambulatory Health Care National Patient Safety Goals. Retrieved from https://www.jointcommission.org/assets/1/6/2019_AHC_NPSGs_final.pdf',
        'Weeks, D. L., Corbette, C. F., & Stream, G. (2010). Beliefs of ambulatory care physicians about accuracy of patient medication records and technology-enhanced solutions to improve accuracy. Journal for Healthcare Quality, 32(5), 12-21. doi:10.1111/j.1945-1474.2010.00097.x',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: All visits occurring during the 12 month measurement period for
        patients aged 18 years and older
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: None

        Exceptions: Documentation of a medical reason(s) for not documenting, updating, or
        reviewing the patient’s current medications list (e.g., patient is in an urgent or emergent
        medical situation where time is of the essence and to delay treatment would jeopardize the
        patient's health status)
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Eligible professional or eligible clinician attests to documenting, updating or
        reviewing the patient's current medications using all immediate resources available on the
        date of the encounter

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: The Joint Commission's 2019 Ambulatory Health Care National
        Patient Safety Goals guide providers to maintain and communicate accurate patient
        medication information. Specifically, the section "Use Medicines Safely NPSG.03.06.01"
        states the following: "Record and pass along correct information about a patient’s
        medicines. Find out what medicines the patient is taking. Compare those medicines to new
        medicines given to the patient. Make sure the patient knows which medicines to take when
        they are at home. Tell the patient it is important to bring their up-to-date list of
        medicines every time they visit a doctor."

        The National Quality Forum's Safe Practices for Better Healthcare - 2010 Update, states the
        following: "the healthcare organization must develop, reconcile, and communicate an
        accurate patient medication list throughout the continuum of care" (p. 40).
        """
        pass



from enum import Enum

COMMAND_TYPE_ALLERGY = 'allergy'
COMMAND_TYPE_DIAGNOSE = 'diagnose'
COMMAND_TYPE_FOLLOWUP = 'followUp'
COMMAND_TYPE_IMAGING = 'imagingOrder'
COMMAND_TYPE_IMMUNIZATION = 'immunization'
COMMAND_TYPE_INSTRUCTION = 'instruct'
COMMAND_TYPE_INTERVIEW = 'interview'
COMMAND_TYPE_LAB_ORDER = 'labOrder'
COMMAND_TYPE_PERFORM = 'perform'
COMMAND_TYPE_PLAN = 'plan'
COMMAND_TYPE_PRESCRIBE = 'prescribe'
COMMAND_TYPE_REFERRAL = 'refer'
COMMAND_TYPE_STRUCTURED_ASSESSMENT = 'structuredAssessment'
COMMAND_TYPE_TASK = 'task'
COMMAND_TYPE_VITALSIGN = 'vitalSign'


class CHANGE_TYPE:

    """
    Change types tell the system when to recompute this protocol.
    Include only the types that are used to compute whether the protocol is used to
    ensure the system processes data in an optimized manner.

    ```python
    class ClinicalQualityMeasure122v6(ClinicalQualityMeasure):

        ...

        compute_on_change_types = [
            CHANGE_TYPE.PROTOCOL_OVERRIDE,
            CHANGE_TYPE.CONDITION,
            CHANGE_TYPE.LAB_REPORT,
            CHANGE_TYPE.PATIENT,
        ]
    ```
    """

    ALLERGY_INTOLERANCE = 'allergy_intolerance'
    "Recompute protocol on an update to allergy information."

    APPOINTMENT = 'appointment'
    "Recompute protocol when a patient's appointment is updated."

    BILLING_LINE_ITEM = 'billing_line_item'
    "Recompute protocol when billing line items are updated."

    CONDITION = 'condition'
    "Recompute protocol when a patient's condition is updated."

    CONSENT = 'consent'
    "Recompute protocol when a patient's consents are updated"

    COVERAGE = 'coverage'
    "Recompute protocol when a patient's insurance coverage is updated."

    ENCOUNTER = 'encounter'
    "Recompute protocol when a patient's insurance coverage is updated."

    IMAGING_REPORT = 'imaging_report'
    "Recompute protocol when a patient's imaging report is updated."

    IMMUNIZATION = 'immunization'
    "Recompute protocol when a patient's immunization records have been updated."

    INSTRUCTION = 'instruction'
    "Recompute protocol when a patient's instruction has been updated."

    INTERVIEW = 'interview'
    "Recompute protocol when a patient's interview record has been updated."

    LAB_REPORT = 'lab_report'
    "Recompute protocol when a lab order has been updated."

    MEDICATION = 'medication'
    "Recompute protocol when a medication has been updated."

    MESSAGE = 'message'
    "Recompute protocol when a patient message transmission has been created."

    PATIENT = 'patient'
    "Recompute protocol when a patient's demographic information has been updated."

    PRESCRIPTION = 'prescription'
    "Recompute protocol when a patient's prescription information has been updated"

    PROCEDURE = 'procedure'
    "Recompute protocol when a procedure is performed on a patient"

    PROTOCOL_OVERRIDE = 'protocol_override'
    "Recompute protocol when a protocol is overridden."

    REFERRAL_REPORT = 'referral_report'
    "Recompute protocol when a referral or referral report is updated."

    SUSPECT_HCC = 'suspect_hcc'
    "Recompute protocol when hierarchical condition categories are updated."

    TASK = 'task'
    "Recompute protocol when a task related to the patient is updated"

    VITAL_SIGN = 'vital_sign'
    "Recompute protocol when a patient's vital signs are updated."


class AlertPlacement(Enum):
    ALERT_PLACEMENT_CHART = 'chart'
    ALERT_PLACEMENT_TIMELINE = 'timeline'
    ALERT_PLACEMENT_APPOINTMENT_CARD = 'appointment_card'
    ALERT_PLACEMENT_SCHEDULING_CARD = 'scheduling_card'
    ALERT_PLACEMENT_PROFILE = 'profile'


class AlertIntent(Enum):
    ALERT_INTENT_INFO = 'info'
    ALERT_INTENT_WARNING = 'warning'
    ALERT_INTENT_ALERT = 'alert'



from typing import Dict, List

import arrow

from cached_property import cached_property

from canvas_workflow_kit import events
from canvas_workflow_kit.protocol import (
    STATUS_DUE,
    STATUS_SATISFIED,
    ClinicalQualityMeasure,
    ProtocolResult
)
from canvas_workflow_kit.recommendation import Recommendation
from canvas_workflow_kit.value_set.hcc2018 import HCCConditions


class Hcc001v1(ClinicalQualityMeasure):

    class Meta:
        title = 'Problem List Hygiene'
        version = '2019-02-12v1'
        description = 'All patients with active condition not assessed within the last year.'
        information = 'https://canvas-medical.zendesk.com/hc/en-us/articles/360059083693-Problem-List-Hygiene-HCC001v1'  # noqa: E501

        identifiers = ['HCC001v1']

        types = ['HCC']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]
        authors = [
            'Canvas Medical Team',
        ]

        references = [
            'Canvas Medical HCC, https://canvas-medical.zendesk.com/hc/en-us/articles/360059083693-Problem-List-Hygiene-HCC001v1'  # noqa: E501
        ]

        compute_on_change_types = [
            ClinicalQualityMeasure.CHANGE_PROTOCOL_OVERRIDE,
            ClinicalQualityMeasure.CHANGE_CONDITION,
        ]

    @cached_property
    def active_hcc(self) -> List[Dict]:
        hcc_conditions = []
        for record in [
                r for r in self.patient.conditions.find(HCCConditions)
                if r['clinicalStatus'] == 'active'
        ]:
            if record['lastTimestamps']['assessed']:
                last_date = arrow.get(record['lastTimestamps']['assessed'])
            else:
                last_date = arrow.get(record['noteTimestamp'])

            codes = [c for c in record['coding'] if c['system'] == 'ICD-10']
            if codes:
                hcc_conditions.append({
                    'ICD10': codes[0]['code'],
                    'date': last_date,
                })
        return hcc_conditions

    @cached_property
    def too_old_hccs(self) -> List:
        return [hcc for hcc in self.active_hcc if hcc['date'] < self.timeframe.start]

    def in_initial_population(self) -> bool:
        return True

    def in_denominator(self) -> bool:
        """
        Patients with active condition in the HCC list
        """
        return bool(self.active_hcc)

    def in_numerator(self) -> bool:
        """
        Patients with active condition in the HCC list
         did not have a prior assessment or diagnosis in past 12 months
        """
        return bool(self.too_old_hccs)

    def compute_results(self) -> ProtocolResult:
        result = ProtocolResult()
        if self.in_denominator():
            if self.in_numerator():
                result.due_in = -1
                result.status = STATUS_DUE
                for hcc in self.too_old_hccs:
                    idc10 = hcc['ICD10']
                    label = HCCConditions.label_idc10_for(idc10)
                    raf = HCCConditions.raf_for(idc10)
                    day = hcc['date'].format('M/D/YY')
                    result.add_narrative(
                        f'{label} ({idc10}) is a significant condition which should be assessed annually. '  # noqa: E501
                        f'The condition was last assessed on {day} and carries a RAF value of {raf}'
                    )

                result.add_recommendation(
                    Recommendation(
                        key='HCC001v1_RECOMMEND_ASSESS_CONDITION',
                        rank=1,
                        button='Assess',
                        title='Assess, update or resolve conditions as clinically appropriate',
                        narrative=(
                            '{0} has un-assessed HCC conditions for more than one year'.format(
                                self.patient.first_name)),
                        command={'key': 'assess'}))
                result.add_recommendation(
                    Recommendation(
                        key='HCC001v1_RECOMMEND_RESOLVE_CONDITION',
                        rank=2,
                        button='Assess',
                        title='Resolve conditions as clinically appropriate',
                        narrative=(
                            '{0} has un-assessed HCC conditions for more than one year'.format(
                                self.patient.first_name)),
                        command={'key': 'resolveCondition'}))
            else:
                result.due_in = (min([hcc['date'] for hcc in self.active_hcc
                                      ]).shift(days=self.timeframe.duration) - self.now).days
                result.status = STATUS_SATISFIED
                result.add_narrative('All Significant Condition have been assessed within the last {0}.'.format(  # noqa: E501
                    self.now.shift(months=-1, days=-1 * self.timeframe.duration)
                        .humanize(other=self.now, granularity='month', only_distance=True)
                ).replace(' ago', ''))  # yapf: disable
        return result



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    AndrogenDeprivationTherapyForUrologyCare,
    DexaDualEnergyXrayAbsorptiometryBoneDensityForUrologyCare, Ethnicity, OfficeVisit,
    OncAdministrativeSex, Payer, ProstateCancer, Race)


class ClinicalQualityMeasure645v4(ClinicalQualityMeasure):
    """
    Bone density evaluation for patients with prostate cancer and receiving androgen deprivation
    therapy

    Description: Patients determined as having prostate cancer who are currently starting or
    undergoing androgen deprivation therapy (ADT), for an anticipated period of 12 months or
    greater and who receive an initial bone density evaluation. The bone density evaluation must be
    prior to the start of ADT or within 3 months of the start of ADT.

    Definition: Data Criteria Value Set 2.16.840.1.113762.1.4.1151.38 DEXA, Dual Energy Xray
    Absorptiometry, Bone Density for Urology Care contains 2 LOINC codes identifying the axial and
    appendicular skeleton and will meet the measure intent.

    DEXA - Dual Energy X-ray Absorptiometry - A scan that measures the bone of the spine, hip or
    total body and measures bone mineral density. It is considered one of the most accurate
    measurements.
    PDXA - Peripheral Dual Energy X-ray Absorptiometry - Bone mineral density measurement of the
    wrist, heel or finger.

    First Androgen Deprivation Therapy - The First Androgen Deprivation Therapy (ADT) is measured
    as the first order or administration of ADT for an anticipated period of 12 months or greater
    to a patient with prostate cancer.

    Rationale: Androgen suppression as a treatment for prostate cancer can cause osteoporosis
    (Qaseem, 2008). Men undergoing prolonged androgen deprivation therapy (ADT) incur bone loss at
    a rate higher than menopausal women (Guise, 2007). In preserving bone health, the goal is to
    prevent or treat osteopenia/osteoporosis for the patient on ADT and to prevent or delay
    skeletal related events. The National Osteoporosis Foundation recommendations including a
    baseline assessment of bone density with a DEXA scan and daily calcium and Vitamin D
    supplementation (Watts, 2012). The DEXA scan is the gold standard for bone density screening.
    Men at risk for adverse bone consequences from chronic ADT do not always receive care according
    to evidence-based guidelines. These findings call for improved processes that standardize
    evidence-based practice including baseline and follow up bone density assessment (Watts, 2012).

    Guidance: In order to capture the practitioner's intent of androgen deprivation therapy (ADT)
    for a period of 12 months or greater, SNOMEDCT 456381000124102 which is Injection of leuprolide
    acetate for twelve month period (regime/therapy) is the correct code.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5.  Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms645v4
    """

    title = 'Bone density evaluation for patients with prostate cancer and receiving androgen deprivation therapy'

    identifiers = ['CMS645v4']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'Oregon Urology',
    ]

    references = [
        'Cosman, F., deBeur, S., LeBoff, M., et al. (2014, June). Clinician’s guide to prevention and treatment of osteoporosis. Osteoporosis International, 25(10), 2359–2381. Retrieved from http://link.springer.com/article/10.1007/s00198-014-2794-2',
        'Finkelstein, J., & Yu, E. (2017). Clinical manifestations, diagnosis, and evaluation of osteoporosis in men. Retrieved from https://www.uptodate.com/contents/clinical-manifestations-diagnosis-and-evaluation-of-osteoporosis-in-men/print?source=see_link',
        'Guise, T., Oefelein, M., Eastham, J., et al. (2007). Estrogenic side effects of androgen deprivation therapy. Reviews in Urology, 9(4), 163-180. Retrieved from https://www.researchgate.net/profile/Celestia_Higano/publication/5619579_Estrogenic_side_effects_of_androgen_deprivation_therapy/links/0c960526434483fdfd000000.pdf',
        'Qaseem, A., Snow, V., Shekelle, P., et al. (2008). Annals of Internal Medicine. Screening for osteoporosis in men: A clinical practice guideline from the American College of Physicians, 2008. Annals of Internal Medicine, 148(9), 680-684. Retrieved from http://annals.org/aim/article/740825/screening-osteoporosis-men-clinical-practice-guideline-from-american-college-physicians',
        'Ward, R., Roberts, C., Bencardino, J., et al. (2016). American College of Radiology: ACR Appropriateness Criteria (R)—Osteoporosis and bone mineral density. Retrieved from https://www.jacr.org/article/S1546-1440(17)30198-9/pdf',
        'Watts, N., Adler, R., Bilezikian, J., et al. (2012, June). Osteoporosis in men: An Endocrine Society clinical practice guideline, Journal of Clinical Endocrinology & Metabolism, 97(6), 1802–1822. Retrieved from https://academic.oup.com/jcem/article-lookup/doi/10.1210/jc.2011-3045',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Male patients with a qualifying encounter in the measurement period AND
        with a diagnosis of prostate cancer AND with an order for ADT or an active medication of
        ADT with an intent for treatment greater than or equal to 12 months during the measurement
        period
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: None

        Exceptions: Patient refused recommendation for a bone density evaluation after the start of
        ADT therapy
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Patients with a bone density evaluation within the two years prior to the start
        of or less than three months after the start of ADT treatment

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: Bone density screening should be performed at the start of
        Androgen Deprivation Therapy (ADT) for prostate cancer. It should also be performed every 2
        years for the patient with continued ADT or for patients with known osteoporosis. Current
        insurance practice is to possibly cover the cost of bone density screening if osteoporosis
        is known or if there is a high-risk drug. Some patients choose to delay bone density
        screening until after ADT is started and they therefore have insurance authorization due to
        the administration of a high-risk drug.
        """
        pass
